import logging
import pathlib
from typing import List, Optional, Union

import click
from click import Context

from copernicusmarine.command_line_interface.command_subset_split_on import (
    split_on,
)
from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.command_line_interface.utils import (
    MutuallyExclusiveOption,
    assert_cli_args_are_not_set_except_create_template,
    credentials_file_option,
    force_dataset_part_option,
    force_dataset_version_option,
    force_download_option,
    tqdm_disable_option,
)
from copernicusmarine.core_functions import documentation_utils
from copernicusmarine.core_functions.click_custom_class import (
    CustomClickOptionsGroup,
    CustomDeprecatedClickOption,
)
from copernicusmarine.core_functions.fields_query_builder import (
    build_query,
    get_queryable_requested_fields,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_COORDINATES_SELECTION_METHODS,
    DEFAULT_FILE_FORMATS,
    DEFAULT_VERTICAL_AXES,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    FileFormat,
    ResponseSubset,
    VerticalAxis,
)
from copernicusmarine.core_functions.request_structure import (
    create_subset_request,
)
from copernicusmarine.core_functions.subset import (
    create_subset_template,
    subset_function,
)

logger = logging.getLogger("copernicusmarine")
blank_logger = logging.getLogger("copernicusmarine_blank_logger")

DEFAULT_FIELDS_TO_INCLUDE = {
    "status",
    "message",
    "file_size",
    "data_transfer_size",
}


@click.group()
def cli_subset():
    """Subset command group holder (for CommandCollection)."""
    pass


@click.group(
    "subset",
    cls=CustomClickOptionsGroup,
    short_help="Download subsets of datasets as NetCDF files or Zarr stores.",
    help=documentation_utils.SUBSET["SUBSET_DESCRIPTION_HELP"]
    + "See :ref:`describe <cli-describe>`."
    + " \n\nReturns\n "
    + documentation_utils.SUBSET["SUBSET_RESPONSE_HELP"],
    epilog="""
    Examples:

    .. code-block:: bash

        copernicusmarine subset --dataset-id cmems_mod_ibi_phy_my_0.083deg-3D_P1D-m --variable thetao --variable so --start-datetime 2021-01-01 --end-datetime 2021-01-03 --minimum-longitude 0.0 --maximum-longitude 0.1 --minimum-latitude 28.0 --maximum-latitude 28.1 --minimum-depth 1 --maximum-depth 2

    Equivalent to:

    .. code-block:: bash

        copernicusmarine subset -i cmems_mod_ibi_phy_my_0.083deg-3D_P1D-m -v thetao -v so -t 2021-01-01 -T 2021-01-03 -x 0.0 -X 0.1 -y 28.0 -Y 28.1 -z 1 -Z 2 \n
    """,  # noqa
    invoke_without_command=True,
)
@click.option(
    "--dataset-id",
    "-i",
    type=str,
    default=None,
    help=documentation_utils.SUBSET["DATASET_ID_HELP"],
)
@force_dataset_version_option
@force_dataset_part_option
@click.option(
    "--username",
    type=str,
    default=None,
    help=documentation_utils.SUBSET["USERNAME_HELP"],
)
@click.option(
    "--password",
    type=str,
    default=None,
    help=documentation_utils.SUBSET["PASSWORD_HELP"],
)
@click.option(
    "--variable",
    "-v",
    "variables",
    type=str,
    help=documentation_utils.SUBSET["VARIABLES_HELP"],
    multiple=True,
)
@click.option(
    "--minimum-longitude",
    type=float,
    help=documentation_utils.SUBSET["MINIMUM_LONGITUDE_HELP"],
)
@click.option(
    "-x",
    "alias_min_x",
    type=float,
    help=documentation_utils.SUBSET["ALIAS_MIN_X_HELP"],
)
@click.option(
    "--minimum-x",
    type=float,
    help=documentation_utils.SUBSET["MINIMUM_X_HELP"],
)
@click.option(
    "--maximum-longitude",
    type=float,
    help=documentation_utils.SUBSET["MAXIMUM_LONGITUDE_HELP"],
)
@click.option(
    "-X",
    "alias_max_x",
    type=float,
    help=documentation_utils.SUBSET["ALIAS_MAX_X_HELP"],
)
@click.option(
    "--maximum-x",
    type=float,
    help=documentation_utils.SUBSET["MAXIMUM_X_HELP"],
)
@click.option(
    "--minimum-latitude",
    type=click.FloatRange(min=-90, max=90),
    help=documentation_utils.SUBSET["MINIMUM_LATITUDE_HELP"],
)
@click.option(
    "-y",
    "alias_min_y",
    type=float,
    help=documentation_utils.SUBSET["ALIAS_MIN_Y_HELP"],
)
@click.option(
    "--minimum-y",
    type=float,
    help=documentation_utils.SUBSET["MINIMUM_Y_HELP"],
)
@click.option(
    "--maximum-latitude",
    type=click.FloatRange(min=-90, max=90),
    help=documentation_utils.SUBSET["MAXIMUM_LATITUDE_HELP"],
)
@click.option(
    "-Y",
    "alias_max_y",
    type=float,
    help=documentation_utils.SUBSET["ALIAS_MAX_Y_HELP"],
)
@click.option(
    "--maximum-y",
    type=float,
    help=documentation_utils.SUBSET["MAXIMUM_Y_HELP"],
)
@click.option(
    "--minimum-depth",
    "-z",
    type=float,
    help=documentation_utils.SUBSET["MINIMUM_DEPTH_HELP"],
)
@click.option(
    "--maximum-depth",
    "-Z",
    type=float,
    help=documentation_utils.SUBSET["MAXIMUM_DEPTH_HELP"],
)
@click.option(
    "--vertical-axis",
    "-V",
    type=click.Choice(DEFAULT_VERTICAL_AXES),
    default=DEFAULT_VERTICAL_AXIS,
    help=documentation_utils.SUBSET["VERTICAL_AXIS_HELP"],
)
@click.option(
    "--start-datetime",
    "-t",
    type=str,
    help=documentation_utils.SUBSET["START_DATETIME_HELP"]
    + " Caution: encapsulate date with “ “ to ensure valid "
    "expression for format “%Y-%m-%d %H:%M:%S”.",
)
@click.option(
    "--end-datetime",
    "-T",
    type=str,
    help=documentation_utils.SUBSET["END_DATETIME_HELP"]
    + " Caution: encapsulate date with “ “ to ensure valid "
    "expression for format “%Y-%m-%d %H:%M:%S”.",
)
@click.option(
    "--platform-id",
    "-p",
    "platform_ids",
    type=str,
    help=documentation_utils.SUBSET["PLATFORM_IDS_HELP"],
    multiple=True,
)
@click.option(
    "--coordinates-selection-method",
    type=click.Choice(DEFAULT_COORDINATES_SELECTION_METHODS),
    default=DEFAULT_COORDINATES_SELECTION_METHOD,
    help=documentation_utils.SUBSET["COORDINATES_SELECTION_METHOD_HELP"],
)
@click.option(
    "--output-directory",
    "-o",
    type=click.Path(path_type=pathlib.Path),
    help=documentation_utils.SUBSET["OUTPUT_DIRECTORY_HELP"],
)
@credentials_file_option
@click.option(
    "--output-filename",
    "-f",
    type=str,
    help=documentation_utils.SUBSET["OUTPUT_FILENAME_HELP"],
)
@click.option(
    "--file-format",
    type=click.Choice(DEFAULT_FILE_FORMATS),
    default=None,
    help=documentation_utils.SUBSET["FILE_FORMAT_HELP"],
)
@click.option(
    "--overwrite",
    is_flag=True,
    default=False,
    cls=MutuallyExclusiveOption,
    help=documentation_utils.SUBSET["OVERWRITE_HELP"],
    mutually_exclusive=["skip-existing"],
)
@click.option(
    "--skip-existing",
    is_flag=True,
    type=bool,
    default=False,
    cls=MutuallyExclusiveOption,
    help=documentation_utils.SUBSET["SKIP_EXISTING_HELP"],
    mutually_exclusive=["overwrite"],
)
@click.option(
    "--service",
    "-s",
    type=str,
    help=documentation_utils.SUBSET["SERVICE_HELP"],
)
@click.option(
    "--create-template",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.SUBSET["CREATE_TEMPLATE_HELP"],
)
@click.option(
    "--request-file",
    type=click.Path(exists=True, path_type=pathlib.Path),
    help=documentation_utils.SUBSET["REQUEST_FILE_HELP"],
)
@click.option(
    "--motu-api-request",
    type=str,
    help=documentation_utils.SUBSET["MOTU_API_REQUEST_HELP"],
    cls=CustomDeprecatedClickOption,
    custom_deprecated=["--motu-api-request"],
)
@click.option(
    "--dry-run",
    type=bool,
    is_flag=True,
    default=False,
    help=documentation_utils.SUBSET["DRY_RUN_HELP"],
)
@click.option(
    "--response-fields",
    "-r",
    type=str,
    default=None,
    help=documentation_utils.GET["RESPONSE_FIELDS_HELP"],
)
@click.option(
    "--netcdf-compression-level",
    type=click.IntRange(0, 9),
    is_flag=False,
    flag_value=1,
    default=0,
    help=documentation_utils.SUBSET["NETCDF_COMPRESSION_LEVEL_HELP"]
    + " If used as a flag, the assigned value will be 1.",
)
@click.option(
    "--netcdf3-compatible",
    type=bool,
    default=False,
    is_flag=True,
    help=documentation_utils.SUBSET["NETCDF3_COMPATIBLE_HELP"],
)
@click.option(
    "--chunk-size-limit",
    type=click.IntRange(min=-1),
    default=-1,
    help=documentation_utils.SUBSET["CHUNK_SIZE_LIMIT_HELP"],
)
@click.option(
    "--staging",
    type=bool,
    default=False,
    is_flag=True,
    hidden=True,
)
@tqdm_disable_option
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=documentation_utils.SUBSET["LOG_LEVEL_HELP"],
)
@click.option(
    "--raise-if-updating",
    type=bool,
    default=False,
    is_flag=True,
    help=documentation_utils.SUBSET["RAISE_IF_UPDATING_HELP"],
)
@force_download_option
@click.pass_context
@log_exception_and_exit
def subset(
    context: Context,
    dataset_id: str,
    dataset_version: Optional[str],
    dataset_part: Optional[str],
    username: Optional[str],
    password: Optional[str],
    variables: Optional[List[str]],
    minimum_longitude: Optional[float],
    maximum_longitude: Optional[float],
    minimum_latitude: Optional[float],
    maximum_latitude: Optional[float],
    minimum_x: Optional[float],
    maximum_x: Optional[float],
    minimum_y: Optional[float],
    maximum_y: Optional[float],
    alias_min_x: Optional[float],
    alias_max_x: Optional[float],
    alias_min_y: Optional[float],
    alias_max_y: Optional[float],
    minimum_depth: Optional[float],
    maximum_depth: Optional[float],
    vertical_axis: VerticalAxis,
    start_datetime: Optional[str],
    end_datetime: Optional[str],
    platform_ids: Optional[List[str]],
    coordinates_selection_method: CoordinatesSelectionMethod,
    output_filename: Optional[str],
    file_format: Optional[FileFormat],
    netcdf_compression_level: int,
    netcdf3_compatible: bool,
    service: Optional[str],
    create_template: bool,
    request_file: Optional[pathlib.Path],
    output_directory: Optional[pathlib.Path],
    credentials_file: Optional[pathlib.Path],
    motu_api_request: Optional[str],
    overwrite: bool,
    skip_existing: bool,
    dry_run: bool,
    response_fields: Optional[str],
    disable_progress_bar: bool,
    log_level: str,
    chunk_size_limit: int,
    staging: bool,
    raise_if_updating: bool,
    force_download: bool,
):
    if context.meta["help_for"] == "split-on":
        return
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUG mode activated")

    if create_template:
        assert_cli_args_are_not_set_except_create_template(context)
        create_subset_template()
        return

    subset_request = create_subset_request(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        minimum_depth=minimum_depth,
        maximum_depth=maximum_depth,
        vertical_axis=vertical_axis,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        platform_ids=platform_ids,
        coordinates_selection_method=coordinates_selection_method,
        staging=staging,
        output_filename=output_filename,
        file_format=file_format,
        service=service,
        request_file=request_file,
        output_directory=output_directory,
        credentials_file=credentials_file,
        motu_api_request=motu_api_request,
        overwrite=overwrite,
        skip_existing=skip_existing,
        dry_run=dry_run,
        disable_progress_bar=disable_progress_bar,
        netcdf_compression_level=netcdf_compression_level,
        netcdf3_compatible=netcdf3_compatible,
        chunk_size_limit=chunk_size_limit,
        raise_if_updating=raise_if_updating,
        minimum_longitude=minimum_longitude,
        maximum_longitude=maximum_longitude,
        minimum_latitude=minimum_latitude,
        maximum_latitude=maximum_latitude,
        minimum_x=minimum_x,
        maximum_x=maximum_x,
        minimum_y=minimum_y,
        maximum_y=maximum_y,
        alias_min_x=alias_min_x,
        alias_max_x=alias_max_x,
        alias_min_y=alias_min_y,
        alias_max_y=alias_max_y,
    )

    if context.invoked_subcommand is not None:
        context.ensure_object(dict)
        context.obj["subset_request"] = subset_request
        context.obj["response_fields"] = response_fields

        return

    responses = subset_function(subset_request)

    if response_fields:
        fields_to_include = set(response_fields.replace(" ", "").split(","))
    elif dry_run:
        fields_to_include = {"all"}
    else:
        fields_to_include = DEFAULT_FIELDS_TO_INCLUDE

    included_fields: Optional[Union[dict, set]]
    if "all" in fields_to_include:
        included_fields = None
    elif "none" in fields_to_include:
        included_fields = set()
    else:
        queryable_fields = get_queryable_requested_fields(
            fields_to_include, ResponseSubset, "--response-fields"
        )
        included_fields = build_query(set(queryable_fields), ResponseSubset)

    blank_logger.info(
        responses.model_dump_json(
            indent=2,
            include=included_fields,
            exclude_none=True,
            exclude_unset=True,
        )
    )


subset.add_command(split_on)

cli_subset.add_command(subset)



import pathlib
from datetime import datetime
from typing import List, Optional, Union

import pandas as pd
import xarray

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    deprecated_python_option,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_COORDINATES_SELECTION_METHOD,
    DEFAULT_VERTICAL_AXIS,
    CoordinatesSelectionMethod,
    VerticalAxis,
)
from copernicusmarine.core_functions.open_dataset import open_dataset_function
from copernicusmarine.core_functions.request_structure import (
    create_subset_request,
)
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def open_dataset(
    dataset_id: str,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    maximum_x: Optional[float] = None,
    minimum_x: Optional[float] = None,
    maximum_y: Optional[float] = None,
    minimum_y: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_axis: VerticalAxis = DEFAULT_VERTICAL_AXIS,
    start_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    end_datetime: Optional[Union[datetime, pd.Timestamp, str]] = None,
    coordinates_selection_method: CoordinatesSelectionMethod = (
        DEFAULT_COORDINATES_SELECTION_METHOD
    ),
    service: Optional[str] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    raise_if_updating: bool = False,
    chunk_size_limit: int = -1,
    staging: bool = False,
) -> xarray.Dataset:
    """
    Load an xarray dataset using 'lazy-loading' mode from a Copernicus Marine data source.

    This means that data is only loaded into memory when a computation is called, optimizing RAM usage by avoiding immediate loading.
    It supports various parameters for customization, such as specifying geographical bounds, temporal range, depth range, and more.


    Parameters
    ----------
    dataset_id : str
        The datasetID, required.
    dataset_version : str, optional
        Force the selection of a specific dataset version.
    dataset_part : str, optional
        Force the selection of a specific dataset part.
    username : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    password : str, optional
        If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, then search for a credentials file, else ask for user input. See also :func:`~copernicusmarine.login`
    variables : List[str], optional
        List of variable names to extract.
    minimum_longitude : float, optional
        Minimum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    maximum_longitude : float, optional
        Maximum longitude for the subset. The value will be transposed to the interval [-180; 360[.
    minimum_latitude : float, optional
        Minimum latitude for the subset. Requires a float from -90 degrees to +90.
    maximum_latitude : float, optional
        Maximum latitude for the subset. Requires a float from -90 degrees to +90.
    minimum_x : float, optional
        Minimum x-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_x : float, optional
        Maximum x-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_y : float, optional
        Minimum y-axis value for the subset. The units are considered in length (m, 100km...).
    maximum_y : float, optional
        Maximum y-axis value for the subset. The units are considered in length (m, 100km...).
    minimum_depth : float, optional
        Minimum depth for the subset.
    maximum_depth : float, optional
        Maximum depth for the subset.
    vertical_axis : str, optional
        Consolidate the vertical dimension (the z-axis) as requested: depth with descending positive values, elevation with ascending positive values. Default is depth.
    start_datetime : Union[datetime, str], optional
        The start datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    end_datetime : Union[datetime, str], optional
        The end datetime of the temporal subset. Supports common format parsed by dateutil (https://dateutil.readthedocs.io/en/stable/parser.html).
    coordinates_selection_method : str, optional
        If ``inside``, the selection retrieved will be inside the requested range. If ``strict-inside``, the selection retrieved will be inside the requested range, and an error will be raised if the values don't exist. If ``nearest``, the extremes closest to the requested values will be returned. If ``outside``, the extremes will be taken to contain all the requested interval. The methods ``inside``, ``nearest`` and ``outside`` will display a warning if the request is out of bounds.
    service : str, optional
        Force download through one of the available services using the service name among ['arco-geo-series', 'arco-time-series', 'omi-arco', 'static-arco', 'arco-platform-series'] or its short name among ['geoseries', 'timeseries', 'omi-arco', 'static-arco', 'platformseries'].
    credentials_file : Union[pathlib.Path, str], optional
        Path to a credentials file if not in its default directory (``$HOME/.copernicusmarine``). Accepts .copernicusmarine-credentials / .netrc or _netrc / motuclient-python.ini files.
    raise_if_updating : bool, optional
        If set, raises a :class:`copernicusmarine.DatasetUpdating` error if the dataset is being updated and the subset interval requested overpasses the updating start date of the dataset. Otherwise, a simple warning is displayed.
    chunk_size_limit : int, default -1
        Limit the size of the chunks in the dask array. Default is set to -1 which behaves similarly to 'chunks=auto' from ``xarray``. Positive integer values and '-1' are accepted. This is an experimental feature.

    Returns
    -------
    xarray.Dataset
        The loaded xarray dataset.
    """  # noqa
    if variables is not None:
        _check_type(variables, list, "variables")

    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )

    subset_request = create_subset_request(
        dataset_id=dataset_id,
        dataset_version=dataset_version,
        dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        minimum_longitude=minimum_longitude,
        maximum_longitude=maximum_longitude,
        minimum_latitude=minimum_latitude,
        maximum_latitude=maximum_latitude,
        minimum_depth=minimum_depth,
        maximum_depth=maximum_depth,
        vertical_axis=vertical_axis,
        start_datetime=start_datetime,
        end_datetime=end_datetime,
        minimum_x=minimum_x,
        maximum_x=maximum_x,
        minimum_y=minimum_y,
        maximum_y=maximum_y,
        coordinates_selection_method=coordinates_selection_method,
        service=service,
        credentials_file=(
            pathlib.Path(credentials_file) if credentials_file else None
        ),
        staging=staging,
        chunk_size_limit=chunk_size_limit,
        raise_if_updating=raise_if_updating,
    )
    return open_dataset_function(
        subset_request=subset_request,
    )


def _check_type(value, expected_type: type, name: str):
    if not isinstance(value, expected_type):
        raise TypeError(f"{name} must be of type {expected_type.__name__}")



import logging
import os
import pathlib
import re
from datetime import datetime
from itertools import chain
from typing import Literal, Optional

from botocore.client import ClientError
from dateutil.tz import UTC
from tqdm import tqdm

from copernicusmarine.core_functions.models import (
    FileGet,
    FileStatus,
    ResponseGet,
    S3FileInfo,
    S3FilesDescriptor,
    StatusCode,
    StatusMessage,
)
from copernicusmarine.core_functions.request_structure import (
    GetRequest,
    overload_regex_with_additionnal_filter,
)
from copernicusmarine.core_functions.sessions import (
    get_configured_boto3_session,
)
from copernicusmarine.core_functions.utils import (
    get_unique_filepath,
    parse_access_dataset_url,
    run_concurrently,
    timestamp_parser,
)

logger = logging.getLogger("copernicusmarine")


def download_original_files(
    username: str,
    password: str,
    get_request: GetRequest,
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    create_file_list: Optional[str],
) -> ResponseGet:
    endpoint, bucket, path = parse_access_dataset_url(
        str(get_request.dataset_url)
    )
    if get_request.direct_download:
        files_headers = _download_header_for_direct_download(
            files_to_download=get_request.direct_download,
            endpoint_url=endpoint,
            bucket=bucket,
            path=path,
            sync=get_request.sync,
            directory_out=pathlib.Path(get_request.output_directory),
            username=username,
            no_directories=get_request.no_directories,
            overwrite=get_request.overwrite,
            skip_existing=get_request.skip_existing,
        )
    else:
        files_headers = S3FilesDescriptor(endpoint=endpoint, bucket=bucket)
    if (
        not get_request.direct_download
        or files_headers.files_not_found
        or get_request.regex
    ):
        if files_headers.files_not_found:
            files_not_found_regex = "|".join(
                [
                    re.escape(file_not_found)
                    for file_not_found in files_headers.files_not_found
                ]
            )
            get_request.regex = overload_regex_with_additionnal_filter(
                files_not_found_regex, get_request.regex
            )
        if get_request.index_parts:
            _, _, path = parse_access_dataset_url(
                str(get_request.dataset_url), only_dataset_root_path=True
            )
        files_headers_listing = _download_header(
            endpoint_url=endpoint,
            bucket=bucket,
            path=path,
            regex=get_request.regex,
            username=username,
            sync=get_request.sync,
            create_file_list=create_file_list,
            directory_out=pathlib.Path(get_request.output_directory),
            no_directories=get_request.no_directories,
            skip_existing=get_request.skip_existing,
            overwrite=get_request.overwrite,
            disable_progress_bar=disable_progress_bar,
            only_list_root_path=get_request.index_parts,
        )
        if files_headers_listing.create_file_list is True:
            return ResponseGet(
                files=[],
                files_deleted=None,
                files_not_found=None,
                number_of_files_to_download=0,
                status=StatusCode.FILE_LIST_CREATED,
                message=StatusMessage.FILE_LIST_CREATED,
                total_size=None,
            )
        if files_headers_listing:
            files_headers.endpoint = files_headers_listing.endpoint
            files_headers.bucket = files_headers_listing.bucket
            files_headers.s3_files.extend(files_headers_listing.s3_files)
            files_headers.total_size += files_headers_listing.total_size
            files_headers.files_not_found.extend(
                files_headers_listing.files_not_found
            )

    files_headers = _create_filenames_out(
        files_information=files_headers,
        output_directory=pathlib.Path(get_request.output_directory),
        no_directories=get_request.no_directories,
    )

    if get_request.sync_delete:
        files_headers = _get_files_to_delete_with_sync(
            files_information=files_headers,
            output_directory=pathlib.Path(get_request.output_directory),
        )
        if files_headers.files_to_delete:
            logger.info("Some files will be deleted due to sync delete:")
            for file_to_delete in files_headers.files_to_delete:
                logger.info(file_to_delete)
                file_to_delete.unlink()
    if files_headers.total_size == 0:
        logger.info("No data to download")
        if not files_headers.files_to_delete:
            return create_response_get_from_files_headers(
                files_headers, get_request, "NO_DATA_TO_DOWNLOAD"
            )

    response = create_response_get_from_files_headers(
        files_headers, get_request, "SUCCESS"
    )

    if get_request.dry_run:
        response.status = StatusCode.DRY_RUN
        response.message = StatusMessage.DRY_RUN
        return response
    filenames_in = []
    filenames_out = []
    for s3_file in files_headers.s3_files:
        if not s3_file.ignore:
            filenames_in.append(s3_file.filename_in)
            filenames_out.append(s3_file.filename_out)
    download_files(
        username,
        endpoint,
        bucket,
        filenames_in,
        filenames_out,
        max_concurrent_requests,
        disable_progress_bar,
    )
    return response


def create_response_get_from_files_headers(
    files_headers: S3FilesDescriptor,
    get_request: GetRequest,
    status: Literal["SUCCESS", "DRY_RUN", "NO_DATA_TO_DOWNLOAD"],
) -> ResponseGet:
    endpoint = files_headers.endpoint
    return ResponseGet(
        files=[
            FileGet(
                s3_url=s3_file.filename_in,
                https_url=s3_file.filename_in.replace("s3://", endpoint + "/"),
                file_size=size_to_MB(s3_file.size),
                last_modified_datetime=s3_file.last_modified,
                etag=s3_file.etag,
                output_directory=pathlib.Path(get_request.output_directory),
                filename=s3_file.filename_out.name,
                file_path=s3_file.filename_out,
                file_format=s3_file.filename_out.suffix,
                file_status=FileStatus.get_status(
                    ignore=s3_file.ignore, overwrite=s3_file.overwrite
                ),
            )
            for s3_file in files_headers.s3_files
        ],
        files_deleted=(
            [
                str(file_to_delete)
                for file_to_delete in files_headers.files_to_delete
            ]
            if files_headers.files_to_delete
            else None
        ),
        files_not_found=(
            files_headers.files_not_found
            if files_headers.files_not_found
            else None
        ),
        number_of_files_to_download=len(
            [
                file_to_download
                for file_to_download in files_headers.s3_files
                if not file_to_download.ignore
            ]
        ),
        status=(
            StatusCode.NO_DATA_TO_DOWNLOAD
            if status == "NO_DATA_TO_DOWNLOAD"
            else (
                StatusCode.DRY_RUN
                if status == "DRY_RUN"
                else StatusCode.SUCCESS
            )
        ),
        message=(
            StatusMessage.NO_DATA_TO_DOWNLOAD
            if status == "NO_DATA_TO_DOWNLOAD"
            else (
                StatusMessage.DRY_RUN
                if status == "DRY_RUN"
                else StatusMessage.SUCCESS
            )
        ),
        total_size=size_to_MB(files_headers.total_size),
    )


def _get_files_to_delete_with_sync(
    files_information: S3FilesDescriptor,
    output_directory: pathlib.Path,
) -> S3FilesDescriptor:
    if not files_information.s3_files:
        return files_information
    filenames_out = {
        s3_file.filename_out for s3_file in files_information.s3_files
    }
    product_structure = _local_path_from_s3_url(
        files_information.s3_files[0].filename_in, pathlib.Path("")
    ).parts
    product_id = product_structure[0]
    dataset_id = product_structure[1]
    dataset_level_local_folder = output_directory / product_id / dataset_id

    for local_file in dataset_level_local_folder.glob("**/*"):
        if local_file.is_file() and local_file not in filenames_out:
            files_information.files_to_delete.append(local_file)
    return files_information


def download_files(
    username: str,
    endpoint_url: str,
    bucket: str,
    filenames_in: list[str],
    filenames_out: list[pathlib.Path],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
) -> None:
    for filename_out in filenames_out:
        parent_dir = pathlib.Path(filename_out).parent
        if not parent_dir.is_dir():
            pathlib.Path.mkdir(parent_dir, parents=True)
    if max_concurrent_requests:
        run_concurrently(
            _download_one_file,
            [
                (username, endpoint_url, bucket, in_file, str(out_file))
                for in_file, out_file in zip(
                    filenames_in,
                    filenames_out,
                )
            ],
            max_concurrent_requests,
            tdqm_bar_configuration={
                "disable": disable_progress_bar,
                "desc": "Downloading files",
            },
        )
    else:
        logger.info("Downloading files one by one...")
        with tqdm(
            total=len(filenames_in),
            disable=disable_progress_bar,
            desc="Downloading files",
        ) as pbar:
            for in_file, out_file in zip(filenames_in, filenames_out):
                _download_one_file(
                    username, endpoint_url, bucket, in_file, str(out_file)
                )
                pbar.update(1)


def _download_header(
    endpoint_url: str,
    bucket: str,
    path: str,
    regex: Optional[str],
    username: str,
    sync: bool,
    create_file_list: Optional[str],
    directory_out: pathlib.Path,
    no_directories: bool,
    overwrite: bool,
    skip_existing: bool,
    disable_progress_bar: bool,
    only_list_root_path: bool = False,
) -> S3FilesDescriptor:

    files_headers = S3FilesDescriptor(endpoint=endpoint_url, bucket=bucket)

    raw_filenames = _list_files_on_marine_data_lake_s3(
        username,
        endpoint_url,
        bucket,
        path,
        not only_list_root_path,
        disable_progress_bar,
    )

    for filename, size, last_modified_datetime, etag in raw_filenames:
        if not regex or re.search(regex, filename):
            file_to_append = S3FileInfo(
                filename_in=filename,
                size=float(size),
                last_modified=last_modified_datetime.isoformat(),
                etag=etag,
                ignore=_check_should_be_ignored(
                    filename,
                    size,
                    last_modified_datetime,
                    directory_out,
                    skip_existing,
                    sync,
                    no_directories,
                ),
                overwrite=_check_should_be_overwritten(
                    filename,
                    size,
                    last_modified_datetime,
                    directory_out,
                    sync,
                    overwrite,
                    no_directories,
                ),
            )
            files_headers.add_s3_file(file_to_append)

    if create_file_list and create_file_list.endswith(".txt"):
        download_filename = directory_out / create_file_list
        if not overwrite:
            download_filename = get_unique_filepath(
                directory_out / create_file_list,
            )
        with open(download_filename, "w") as file_out:
            for s3_file in files_headers.s3_files:
                if not s3_file.ignore:
                    file_out.write(f"{s3_file.filename_in}\n")
        files_headers.create_file_list = True
    elif create_file_list and create_file_list.endswith(".csv"):
        download_filename = directory_out / create_file_list
        if not overwrite:
            download_filename = get_unique_filepath(
                directory_out / create_file_list,
            )
        with open(download_filename, "w") as file_out:
            file_out.write("filename,size,last_modified_datetime,etag\n")
            for s3_file in files_headers.s3_files:
                if not s3_file.ignore:
                    file_out.write(
                        f"{s3_file.filename_in},{s3_file.size},"
                        f"{s3_file.last_modified},{s3_file.etag}\n"
                    )
        files_headers.create_file_list = True
    return files_headers


def _download_header_for_direct_download(
    files_to_download: list[str],
    endpoint_url: str,
    bucket: str,
    path: str,
    sync: bool,
    directory_out: pathlib.Path,
    username: str,
    no_directories: bool,
    overwrite: bool,
    skip_existing: bool,
) -> S3FilesDescriptor:

    files_headers = S3FilesDescriptor(endpoint=endpoint_url, bucket=bucket)

    split_path = path.split("/")
    root_folder = split_path[0]
    product_id = split_path[1]
    dataset_id_with_tag = split_path[2]

    for file_to_download in files_to_download:
        file_path = file_to_download.split(f"{dataset_id_with_tag}/")[-1]
        if not file_path:
            logger.warning(
                f"{file_to_download} does not seem to be valid. Skipping."
            )
            files_headers.files_not_found.append(file_to_download)
            continue
        full_path = (
            f"s3://{bucket}/{root_folder}/{product_id}/"
            f"{dataset_id_with_tag}/{file_path}"
        )
        size_last_modified_and_etag = _get_file_size_last_modified_and_etag(
            endpoint_url, bucket, full_path, username
        )
        if size_last_modified_and_etag:
            size, last_modified, etag = size_last_modified_and_etag
            file_to_append = S3FileInfo(
                filename_in=full_path,
                size=size,
                last_modified=last_modified.isoformat(),
                etag=etag,
                ignore=_check_should_be_ignored(
                    full_path,
                    size,
                    last_modified,
                    directory_out,
                    skip_existing,
                    sync,
                    no_directories,
                ),
                overwrite=_check_should_be_overwritten(
                    full_path,
                    size,
                    last_modified,
                    directory_out,
                    sync,
                    overwrite,
                    no_directories,
                ),
            )
            files_headers.add_s3_file(file_to_append)
        else:
            files_headers.files_not_found.append(file_to_download)

    if not files_headers.s3_files:
        logger.warning(
            "No files found to download for direct download. "
            "Please check the files to download. "
            "We will try to list the files available for download "
            "and compare them with the requested files."
        )

    return files_headers


def _check_already_exists(
    filename: str,
    directory_out: pathlib.Path,
    no_directories: bool,
) -> bool:
    filename_out = _create_filename_out(
        filename, directory_out, no_directories
    )
    return filename_out.is_file()


def _check_needs_to_be_synced(
    filename: str,
    size: int,
    last_modified_datetime: datetime,
    directory_out: pathlib.Path,
    no_directories: bool,
) -> bool:
    filename_out = _create_filename_out(
        filename, directory_out, no_directories
    )
    if not filename_out.is_file():
        return True
    else:
        file_stats = filename_out.stat()
        if file_stats.st_size != size:
            return True
        else:
            last_created_datetime_out = timestamp_parser(
                file_stats.st_mtime, unit="s"
            )
            # boto3.s3_resource.Object.last_modified is without microsecond
            # boto3.paginate s3_object["LastModified"] is with microsecond
            last_modified_datetime = last_modified_datetime.replace(
                microsecond=0
            )
            return last_modified_datetime > last_created_datetime_out


def _check_should_be_ignored(
    filename: str,
    size: int,
    last_modified_datetime: datetime,
    directory_out: pathlib.Path,
    skip_existing: bool,
    sync: bool,
    no_directories: bool,
) -> bool:
    return (
        skip_existing
        and _check_already_exists(filename, directory_out, no_directories)
    ) or (
        sync
        and not _check_needs_to_be_synced(
            filename,
            size,
            last_modified_datetime,
            directory_out,
            no_directories,
        )
    )


def _check_should_be_overwritten(
    filename: str,
    size: int,
    last_modified_datetime: datetime,
    directory_out: pathlib.Path,
    sync: bool,
    overwrite: bool,
    no_directories: bool,
) -> bool:
    return (
        overwrite
        and _check_already_exists(filename, directory_out, no_directories)
        or (
            sync
            and _check_needs_to_be_synced(
                filename,
                size,
                last_modified_datetime,
                directory_out,
                no_directories,
            )
        )
    )


def _local_path_from_s3_url(
    s3_url: str, local_directory: pathlib.Path
) -> pathlib.Path:
    return local_directory / pathlib.Path("/".join(s3_url.split("/")[4:]))


def _list_files_on_marine_data_lake_s3(
    username: str,
    endpoint_url: str,
    bucket: str,
    prefix: str,
    recursive: bool,
    disable_progress_bar: bool,
) -> list[tuple[str, int, datetime, str]]:
    s3_client, _ = get_configured_boto3_session(
        endpoint_url, ["ListObjectsV2", "HeadObject"], username
    )

    if not prefix.endswith("/"):
        try:
            s3_client.head_object(Bucket=bucket, Key=prefix)
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code")
            if error_code == "404" or error_code == "NoSuchKey":
                prefix += "/"
            else:
                raise

    paginator = s3_client.get_paginator("list_objects")
    page_iterator = paginator.paginate(
        Bucket=bucket,
        Prefix=prefix,
        Delimiter="/" if not recursive else "",
    )
    logger.info("Listing files on remote server...")
    s3_objects = chain(
        *map(
            lambda page: page.get("Contents", []),
            tqdm(page_iterator, disable=disable_progress_bar),
        )
    )
    files_already_found: list[tuple[str, int, datetime, str]] = []
    for s3_object in s3_objects:
        files_already_found.append(
            (
                f"s3://{bucket}/" + s3_object["Key"],
                s3_object["Size"],
                s3_object["LastModified"].astimezone(tz=UTC),
                s3_object["ETag"],
            )
        )
    return files_already_found


def _get_file_size_last_modified_and_etag(
    endpoint_url: str, bucket: str, file_in: str, username: str
) -> Optional[tuple[int, datetime, str]]:
    s3_client, _ = get_configured_boto3_session(
        endpoint_url, ["HeadObject"], username
    )

    try:
        s3_object = s3_client.head_object(
            Bucket=bucket,
            Key=file_in.replace(f"s3://{bucket}/", ""),
        )
        return (
            s3_object["ContentLength"],
            s3_object["LastModified"].astimezone(tz=UTC),
            s3_object["ETag"],
        )
    except ClientError as e:
        if "404" in str(e):
            logger.warning(
                f"File {file_in} not found on the server. Skipping."
            )
            return None
        else:
            raise e


def _download_one_file(
    username,
    endpoint_url: str,
    bucket: str,
    file_in: str,
    file_out: str,
) -> None:
    s3_client, s3_resource = get_configured_boto3_session(
        endpoint_url,
        ["GetObject", "HeadObject"],
        username,
        return_ressources=True,
    )
    last_modified_date_epoch = s3_resource.Object(
        bucket, file_in.replace(f"s3://{bucket}/", "")
    ).last_modified.timestamp()

    s3_client.download_file(
        bucket,
        file_in.replace(f"s3://{bucket}/", ""),
        file_out,
    )

    try:
        os.utime(
            file_out, (last_modified_date_epoch, last_modified_date_epoch)
        )
    except PermissionError:
        logger.warning(
            f"Permission to modify the last modified date "
            f"of the file {file_out} is denied."
        )


# /////////////////////////////
# --- Tools
# /////////////////////////////


def _create_filenames_out(
    files_information: S3FilesDescriptor,
    output_directory: pathlib.Path = pathlib.Path("."),
    no_directories=False,
) -> S3FilesDescriptor:
    for s3_file in files_information.s3_files:
        filename_in = s3_file.filename_in
        filename_out = _create_filename_out(
            filename_in,
            output_directory,
            no_directories,
        )
        if not s3_file.overwrite and not s3_file.ignore:
            filename_out = get_unique_filepath(
                filepath=filename_out,
            )
        s3_file.filename_out = filename_out

    return files_information


def _create_filename_out(
    file_path: str,
    output_directory: pathlib.Path,
    no_directories: bool,
):
    if no_directories:
        filename_out = (
            pathlib.Path(output_directory) / pathlib.Path(file_path).name
        )
    else:
        # filename_in: s3://mdl-native-xx/native/<product-id>..
        filename_out = _local_path_from_s3_url(file_path, output_directory)
    return filename_out


def size_to_MB(size: float) -> float:
    return size / 1024**2



import logging
from enum import Enum
from itertools import groupby
from typing import Any, Optional, TypeVar, Union

import pystac
from pydantic import BaseModel
from tqdm import tqdm

from copernicusmarine.catalogue_parser.models import (
    PART_DEFAULT,
    CopernicusMarineCatalogue,
    CopernicusMarineDataset,
    CopernicusMarineProduct,
    DatasetIsNotPartOfTheProduct,
    DatasetItem,
    DatasetNotFound,
    get_version_and_part_from_full_dataset_id,
)
from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_debug,
)
from copernicusmarine.core_functions.marine_datastore_config import (
    CatalogueConfig,
    MarineDataStoreConfig,
)
from copernicusmarine.core_functions.sessions import JsonParserConnection
from copernicusmarine.core_functions.utils import run_concurrently

logger = logging.getLogger("copernicusmarine")


def get_dataset_metadata(
    dataset_id: str,
    marine_datastore_config: MarineDataStoreConfig,
    raise_on_error: bool,
) -> Optional[CopernicusMarineDataset]:
    seen_dataset_links = set()
    dataset_items: list[DatasetItem] = []
    catalogue_errors: list[tuple] = []
    for catalogue in marine_datastore_config.catalogues:
        try:
            with JsonParserConnection() as connection:
                stac_url = catalogue.root_metadata_url
                dataset_product_mapping_url = (
                    catalogue.dataset_product_mapping_url
                )
                product_ids = connection.get_json_file(
                    dataset_product_mapping_url
                ).get(dataset_id)
                if not product_ids:
                    continue
                for product_id in product_ids.split(","):
                    url = f"{stac_url}/{product_id}/product.stac.json"
                    product_json = connection.get_json_file(url)
                    product_collection = pystac.Collection.from_dict(
                        product_json
                    )
                    product_datasets_metadata_links = (
                        product_collection.get_item_links()
                    )
                    digital_object_identifier = (
                        product_collection.extra_fields.get("sci:doi", None)
                        if product_collection.extra_fields
                        else None
                    )
                    datasets_metadata_links = []
                    for (
                        dataset_metadata_link
                    ) in product_datasets_metadata_links:
                        if (
                            dataset_id in dataset_metadata_link.href
                            and dataset_metadata_link.href
                            not in seen_dataset_links
                        ):
                            datasets_metadata_links.append(
                                dataset_metadata_link
                            )
                            seen_dataset_links.add(dataset_metadata_link.href)

                    for link in datasets_metadata_links:
                        url = f"{stac_url}/{product_id}/{link.href}"
                        dataset_json = connection.get_json_file(url)
                        dataset_item = _parse_dataset_json_to_pystac_item(
                            dataset_json
                        )
                        if dataset_item:
                            (
                                parsed_id,
                                parsed_version,
                                parsed_part,
                            ) = get_version_and_part_from_full_dataset_id(
                                dataset_item.id
                            )
                            dataset_items.append(
                                DatasetItem(
                                    url=url,
                                    stac_json=dataset_json,
                                    stac_item=dataset_item,
                                    item_id=dataset_item.id,
                                    parsed_id=parsed_id,
                                    parsed_version=parsed_version,
                                    parsed_part=parsed_part,
                                    product_doi=digital_object_identifier,
                                )
                            )
        except Exception as exception:
            catalogue_errors.append((exception, catalogue.root_metadata_url))
    if len(catalogue_errors) == len(marine_datastore_config.catalogues):
        first_error = catalogue_errors[0][0]
        raise first_error
    else:
        for error, catalogue_root in catalogue_errors:
            logger.debug(
                f"Error while fetching dataset metadata from {catalogue_root}: {error}"
            )

    if not dataset_items:
        raise DatasetNotFound(dataset_id)

    return _parse_and_sort_dataset_items(
        [
            dataset_item
            for dataset_item in dataset_items
            if dataset_item.parsed_id == dataset_id
        ],
        raise_on_error=raise_on_error,
    )


def _parse_dataset_json_to_pystac_item(
    metadate_json: dict,
) -> Optional[pystac.Item]:
    try:
        return pystac.Item.from_dict(metadate_json)
    except pystac.STACError as exception:
        message = (
            "Invalid Item: If datetime is None, a start_datetime "
            + "and end_datetime must be supplied."
        )
        if exception.args[0] != message:
            logger.error(exception)
            raise pystac.STACError(exception.args)
    return None


def _parse_product_json_to_pystac_collection(
    metadata_json: dict,
) -> Optional[pystac.Collection]:
    try:
        return pystac.Collection.from_dict(metadata_json)
    except KeyError as exception:
        messages = ["spatial", "temporal"]
        if exception.args[0] not in messages:
            logger.error(exception)
        return None


def _parse_and_sort_dataset_items(
    dataset_items: list[DatasetItem],
    raise_on_error: bool,
) -> Optional[CopernicusMarineDataset]:
    """
    Return all dataset metadata parsed and sorted.
    The first version and part are the default.
    """
    dataset_without_parts: list[DatasetItem] = []
    for dataset_item in dataset_items:
        if dataset_item.parsed_part == PART_DEFAULT:
            dataset_without_parts.append(dataset_item)
            break
    if dataset_without_parts:
        dataset_item_example = dataset_without_parts[0]
        dataset_id = dataset_item_example.parsed_id
        dataset_title = dataset_item_example.stac_item.properties.get(
            "title", dataset_id
        )

    else:
        dataset_item_example = dataset_items[0]
        dataset_id = dataset_item_example.parsed_id
        dataset_title = dataset_id
    dataset_part_version_merged = CopernicusMarineDataset(
        dataset_id=dataset_id,
        dataset_name=dataset_title,
        digital_object_identifier=dataset_item_example.product_doi,
        versions=[],
    )
    dataset_part_version_merged.parse_dataset_metadata_items(
        dataset_items, raise_on_error=raise_on_error
    )

    if dataset_part_version_merged.versions == []:
        return None

    dataset_part_version_merged.sort_versions()
    return dataset_part_version_merged


def _construct_marine_data_store_product(
    stac_tuple: tuple[pystac.Collection, list[DatasetItem]],
    raise_on_error: bool,
) -> Optional[CopernicusMarineProduct]:
    stac_product, dataset_items = stac_tuple
    stac_datasets_sorted = sorted(dataset_items, key=lambda x: x.item_id)
    dataset_items_by_dataset_id = groupby(
        stac_datasets_sorted,
        key=lambda x: x.parsed_id,
    )

    digital_object_identifier = (
        stac_product.extra_fields.get("sci:doi", None)
        if stac_product.extra_fields
        else None
    )

    datasets = [
        dataset_metadata
        for _, dataset_items in dataset_items_by_dataset_id
        if (
            dataset_metadata := _parse_and_sort_dataset_items(
                [
                    DatasetItem(
                        url=dataset_item.url,
                        stac_json=dataset_item.stac_json,
                        stac_item=dataset_item.stac_item,
                        item_id=dataset_item.item_id,
                        parsed_id=dataset_item.parsed_id,
                        parsed_version=dataset_item.parsed_version,
                        parsed_part=dataset_item.parsed_part,
                        product_doi=digital_object_identifier,
                    )
                    for dataset_item in dataset_items
                ],
                raise_on_error=raise_on_error,
            )
        )
    ]

    production_center = [
        provider.name
        for provider in stac_product.providers or []
        if provider.roles and "producer" in provider.roles
    ]

    production_center_name = production_center[0] if production_center else ""

    thumbnail_url = None
    if stac_product.assets:
        thumbnail = stac_product.assets.get("thumbnail")
        if thumbnail:
            thumbnail_url = thumbnail.get_absolute_href()

    sources = _get_stac_product_property(stac_product, "sources") or []
    processing_level = _get_stac_product_property(
        stac_product, "processingLevel"
    )

    try:
        return CopernicusMarineProduct(
            title=stac_product.title or stac_product.id,
            product_id=stac_product.id,
            thumbnail_url=thumbnail_url or "",
            description=stac_product.description,
            digital_object_identifier=digital_object_identifier,
            sources=sources,
            processing_level=processing_level,
            production_center=production_center_name,
            keywords=stac_product.keywords,
            datasets=datasets,
        )
    except Exception as e:
        logger.debug(
            f"Error while parsing product {stac_product.title}: {e}",
            exc_info=True,
        )
        if raise_on_error:
            raise e
        return None


def _get_stac_product_property(
    stac_product: pystac.Collection, property_key: str
) -> Optional[Any]:
    properties: dict[str, str] = (
        stac_product.extra_fields.get("properties", {})
        if stac_product.extra_fields
        else {}
    )
    return properties.get(property_key)


def fetch_dataset_items(
    root_url: str,
    connection: JsonParserConnection,
    collection: pystac.Collection,
    force_dataset_id: Optional[str],
    raise_on_error: bool,
) -> list[DatasetItem]:
    items: list[DatasetItem] = []
    for link in collection.get_item_links():
        if not link.owner:
            logger.warning(f"Invalid Item, no owner for: {link.href}")
            continue
        if force_dataset_id and force_dataset_id not in link.href:
            continue
        url = root_url + "/" + link.owner.id + "/" + link.href
        try:
            item_json = connection.get_json_file(url)
            item = _parse_dataset_json_to_pystac_item(item_json)
        except Exception as e:
            logger.debug(
                f"Failed to fetch or parse JSON for dataset URL: {url}"
            )
            if raise_on_error:
                raise e
            log_exception_debug(e)
            continue
        if item:
            (
                parsed_id,
                parsed_version,
                parsed_part,
            ) = get_version_and_part_from_full_dataset_id(item.id)
            items.append(
                DatasetItem(
                    url=url,
                    stac_json=item_json,
                    stac_item=item,
                    item_id=item.id,
                    parsed_id=parsed_id,
                    parsed_version=parsed_version,
                    parsed_part=parsed_part,
                    product_doi=None,
                )
            )
    return items


def fetch_collection(
    root_url: str,
    connection: JsonParserConnection,
    url: str,
    force_dataset_id: Optional[str],
    raise_on_error: bool,
) -> Optional[tuple[pystac.Collection, list[DatasetItem]]]:
    try:
        json_collection = connection.get_json_file(url)
        collection = _parse_product_json_to_pystac_collection(json_collection)
    except Exception as e:
        logger.debug(f"Failed to fetch or parse JSON for product URL: {url}")
        if raise_on_error:
            raise e
        log_exception_debug(e)
        return None
    if collection:
        items = fetch_dataset_items(
            root_url, connection, collection, force_dataset_id, raise_on_error
        )
        return (collection, items)
    return None


def fetch_product_items(
    root_url: str,
    connection: JsonParserConnection,
    child_links: list[pystac.Link],
    force_product_id: Optional[str],
    force_dataset_id: Optional[str],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    raise_on_error: bool,
) -> list[Optional[tuple[pystac.Collection, list[DatasetItem]]]]:
    tasks = []
    for link in child_links:
        if force_product_id and force_product_id not in link.href:
            continue
        tasks.append(
            (
                root_url,
                connection,
                link.absolute_href,
                force_dataset_id,
                raise_on_error,
            )
        )
    tdqm_bar_configuration = {
        "desc": "Fetching products",
        "disable": disable_progress_bar,
        "leave": False,
    }
    return [
        result
        for result in run_concurrently(
            fetch_collection,
            tasks,
            max_concurrent_requests,
            tdqm_bar_configuration,
        )
        if result is not None
    ]


def fetch_all_products_items(
    connection: JsonParserConnection,
    force_product_id: Optional[str],
    force_dataset_id: Optional[str],
    max_concurrent_requests: int,
    catalogue_config: CatalogueConfig,
    disable_progress_bar: bool,
    raise_on_error: bool,
) -> list[Optional[tuple[pystac.Collection, list[DatasetItem]]]]:
    catalog_root_url = catalogue_config.stac_catalogue_url
    json_catalog = connection.get_json_file(catalog_root_url)
    catalog = pystac.Catalog.from_dict(json_catalog)
    catalog.set_self_href(catalog_root_url)
    child_links = catalog.get_child_links()
    root_url = catalogue_config.root_metadata_url
    childs = fetch_product_items(
        root_url,
        connection,
        child_links,
        force_product_id,
        force_dataset_id,
        max_concurrent_requests,
        disable_progress_bar,
        raise_on_error,
    )
    return childs


def parse_catalogue(
    force_product_id: Optional[str],
    force_dataset_id: Optional[str],
    max_concurrent_requests: int,
    disable_progress_bar: bool,
    catalogue_config: CatalogueConfig,
    catalogue_number: int,
    raise_on_error: bool,
) -> Optional[CopernicusMarineCatalogue]:
    logger.debug("Parsing catalogue...")
    progress_bar = tqdm(
        total=2,
        desc=f"Fetching catalogue {catalogue_number + 1}",
        disable=disable_progress_bar,
    )
    dataset_product_mapping_url = catalogue_config.dataset_product_mapping_url
    with JsonParserConnection() as connection:
        if force_dataset_id:
            product_id_from_mapping = connection.get_json_file(
                dataset_product_mapping_url
            ).get(force_dataset_id)
            if not product_id_from_mapping:
                raise DatasetNotFound(force_dataset_id)
            if (
                force_product_id
                and product_id_from_mapping != force_product_id
            ):
                raise DatasetIsNotPartOfTheProduct(
                    force_dataset_id, force_product_id
                )
            force_product_id = product_id_from_mapping
        marine_data_store_root_collections = fetch_all_products_items(
            connection=connection,
            force_product_id=force_product_id,
            force_dataset_id=force_dataset_id,
            max_concurrent_requests=max_concurrent_requests,
            catalogue_config=catalogue_config,
            disable_progress_bar=disable_progress_bar,
            raise_on_error=raise_on_error,
        )
    progress_bar.update()

    products_metadata: list[CopernicusMarineProduct] = []
    for product_item in marine_data_store_root_collections:
        if product_item:
            if product_metadata := _construct_marine_data_store_product(
                product_item, raise_on_error
            ):
                if product_metadata.datasets:
                    products_metadata.append(product_metadata)
    if not products_metadata:
        return None

    products_metadata.sort(key=lambda x: x.product_id)

    full_catalog = CopernicusMarineCatalogue(products=products_metadata)

    progress_bar.update()
    logger.debug("Catalogue parsed")
    return full_catalog


# ---------------------------------------
# --- Utils functions
# ---------------------------------------


def search_and_filter(
    model: BaseModel, search_str: set[str]
) -> Union[BaseModel, None]:
    filtered_fields = {}
    search_str = {s.lower() for s in search_str}
    for field, value in model:
        if isinstance(value, BaseModel):
            filtered_value = search_and_filter(value, search_str)
            if filtered_value:
                filtered_fields[field] = filtered_value

        elif isinstance(value, list) or isinstance(value, tuple):
            filtered_list = []
            for item in value:
                if isinstance(item, BaseModel):
                    filtered_item = search_and_filter(item, search_str)
                    if filtered_item:
                        filtered_list.append(filtered_item)
                elif isinstance(item, str) and any(
                    s in item.lower() for s in search_str
                ):
                    filtered_list.append(item)

            if filtered_list and isinstance(value, list):
                filtered_fields[field] = filtered_list

            if filtered_list and isinstance(value, tuple):
                filtered_fields[field] = tuple(filtered_list)

        elif isinstance(value, dict):
            filtered_dict = {}
            for key, val in value.items():
                if isinstance(val, BaseModel):
                    filtered_val = search_and_filter(val, search_str)
                    if filtered_val:
                        filtered_dict[key] = filtered_val
                elif isinstance(val, str) and any(
                    s in val.lower() for s in search_str
                ):
                    filtered_dict[key] = val

            if filtered_dict:
                filtered_fields[field] = filtered_dict

        elif isinstance(value, Enum):
            if any(s in value.name.lower() for s in search_str):
                filtered_fields[field] = value

        elif isinstance(value, str) and any(
            s in value.lower() for s in search_str
        ):
            filtered_fields[field] = value
    if filtered_fields:
        return model.model_copy(update=filtered_fields)
    return None


def filter_catalogue_with_strings(
    catalogue: CopernicusMarineCatalogue, search_str: set[str]
) -> CopernicusMarineCatalogue:
    filtered_models = []
    for model in catalogue.products:
        filtered_model = search_and_filter(model, search_str)
        if filtered_model:
            filtered_models.append(filtered_model)
    return CopernicusMarineCatalogue(products=filtered_models)


def merge_catalogues(
    catalogues: list[CopernicusMarineCatalogue],
) -> CopernicusMarineCatalogue:
    T = TypeVar("T")

    def next_object(iterable: list[T], key: str, value: str) -> T:
        return next(item for item in iterable if getattr(item, key) == value)

    merged_catalogue = CopernicusMarineCatalogue(products=[])
    seen_products = set()
    seen_datasets = set()
    for catalogue in catalogues:
        for product in catalogue.products:
            if product.product_id not in seen_products:
                merged_catalogue.products.append(product)
                seen_products.add(product.product_id)
                for dataset in product.datasets:
                    seen_datasets.add(dataset.dataset_id)
            else:
                for dataset in product.datasets:
                    merge_product = next_object(
                        merged_catalogue.products,
                        "product_id",
                        product.product_id,
                    )
                    if dataset.dataset_id not in seen_datasets:
                        merge_product.datasets.append(dataset)
                        seen_datasets.add(dataset.dataset_id)
                    else:
                        for version in dataset.versions:
                            merge_dataset = next_object(
                                merge_product.datasets,
                                "dataset_id",
                                dataset.dataset_id,
                            )
                            if version.label not in {
                                version.label
                                for version in merge_dataset.versions
                            }:
                                merge_dataset.versions.append(version)
                            else:
                                for part in version.parts:
                                    merge_version = next_object(
                                        merge_dataset.versions,
                                        "label",
                                        version.label,
                                    )
                                    if part.name not in {
                                        part.name
                                        for part in merge_version.parts
                                    }:
                                        merge_version.parts.append(part)

    return merged_catalogue



import pandas as pd

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.exceptions import ServiceNotSupported
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.services_utils import RetrievalService
from copernicusmarine.core_functions.subset import (
    retrieve_metadata_and_check_request,
)
from copernicusmarine.download_functions.download_sparse import (
    read_dataframe_sparse,
)
from copernicusmarine.download_functions.download_zarr import (
    get_dataset_and_parameters,
)


def read_dataframe_function(
    subset_request: SubsetRequest,
) -> pd.DataFrame:
    retrieval_service: RetrievalService = retrieve_metadata_and_check_request(
        subset_request
    )
    if retrieval_service.service_name not in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.PLATFORMSERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        raise ServiceNotSupported(retrieval_service.service_name)
    if (
        retrieval_service.service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ):
        return read_dataframe_sparse(
            username=subset_request.username,
            subset_request=subset_request,
            metadata_url=retrieval_service.metadata_url,
            service=retrieval_service.service,
            product_doi=retrieval_service.product_doi,
            disable_progress_bar=subset_request.disable_progress_bar,
        )
    else:
        dataset, _, _ = get_dataset_and_parameters(
            subset_request=subset_request,
            dataset_url=retrieval_service.uri,
            axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
            service=retrieval_service.service,
            dataset_chunking=retrieval_service.dataset_chunking,
            is_original_grid=retrieval_service.is_original_grid,
            dataset_valid_start_date=retrieval_service.dataset_valid_start_date,
        )
        return dataset.to_dataframe()



import logging
import os

logger = logging.getLogger("copernicusmarine")

COPERNICUSMARINE_SERVICE_USERNAME = os.getenv(
    "COPERNICUSMARINE_SERVICE_USERNAME"
) or os.getenv("COPERNICUS_MARINE_SERVICE_USERNAME")
if os.getenv("COPERNICUS_MARINE_SERVICE_USERNAME"):
    logger.warning(
        "COPERNICUS_MARINE_SERVICE_USERNAME is deprecated. "
        "Please use COPERNICUSMARINE_SERVICE_USERNAME instead."
    )
COPERNICUSMARINE_SERVICE_PASSWORD = os.getenv(
    "COPERNICUSMARINE_SERVICE_PASSWORD"
) or os.getenv("COPERNICUS_MARINE_SERVICE_PASSWORD")
if os.getenv("COPERNICUS_MARINE_SERVICE_PASSWORD"):
    logger.warning(
        "COPERNICUS_MARINE_SERVICE_PASSWORD is deprecated. "
        "Please use COPERNICUSMARINE_SERVICE_PASSWORD instead."
    )

COPERNICUSMARINE_CREDENTIALS_DIRECTORY = os.getenv(
    "COPERNICUSMARINE_CREDENTIALS_DIRECTORY"
)

COPERNICUSMARINE_DISABLE_SSL_CONTEXT = os.getenv(
    "COPERNICUSMARINE_DISABLE_SSL_CONTEXT", "False"
)

COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH = os.getenv(
    "COPERNICUSMARINE_SET_SSL_CERTIFICATE_PATH"
)

COPERNICUSMARINE_TRUST_ENV = os.getenv("COPERNICUSMARINE_TRUST_ENV", "True")

COPERNICUSMARINE_HTTPS_TIMEOUT = os.getenv(
    "COPERNICUSMARINE_HTTPS_TIMEOUT", "60"
)

COPERNICUSMARINE_HTTPS_RETRIES = os.getenv(
    "COPERNICUSMARINE_HTTPS_RETRIES", "5"
)

PROXY_HTTPS = os.getenv("HTTPS_PROXY", "")
PROXY_HTTP = os.getenv("HTTP_PROXY", "")

COPERNICUSMARINE_CONFIG_URL = os.getenv(
    "COPERNICUSMARINE_CONFIG_URL",
)



import base64
import configparser
import logging
import pathlib
from netrc import netrc
from platform import system
from typing import Literal, Optional, Union

import click
import requests

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_CREDENTIALS_DIRECTORY,
    COPERNICUSMARINE_SERVICE_PASSWORD,
    COPERNICUSMARINE_SERVICE_USERNAME,
)
from copernicusmarine.core_functions.sessions import (
    BearerAuth,
    get_configured_requests_session,
)

logger = logging.getLogger("copernicusmarine")

USER_DEFINED_CACHE_DIRECTORY: str = (
    COPERNICUSMARINE_CREDENTIALS_DIRECTORY or ""
)
DEFAULT_CLIENT_BASE_DIRECTORY: pathlib.Path = (
    pathlib.Path(USER_DEFINED_CACHE_DIRECTORY)
    if USER_DEFINED_CACHE_DIRECTORY
    else pathlib.Path.home()
) / ".copernicusmarine"
DEFAULT_CLIENT_CREDENTIALS_FILENAME = ".copernicusmarine-credentials"
DEFAULT_CLIENT_CREDENTIALS_FILEPATH = (
    DEFAULT_CLIENT_BASE_DIRECTORY / DEFAULT_CLIENT_CREDENTIALS_FILENAME
)
RECOVER_YOUR_CREDENTIALS_MESSAGE = (
    "Learn how to recover your credentials at: "
    "https://help.marine.copernicus.eu/en/articles/"
    "4444552-i-forgot-my-username-or-my-password-what-should-i-do"
)

COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN = "auth.marine.copernicus.eu"
COPERNICUS_MARINE_AUTH_SYSTEM_URL = (
    f"https://{COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN}/"
)
COPERNICUS_MARINE_AUTH_SYSTEM_TOKEN_ENDPOINT = (
    COPERNICUS_MARINE_AUTH_SYSTEM_URL
    + "realms/MIS/protocol/openid-connect/token"
)
COPERNICUS_MARINE_AUTH_SYSTEM_USERINFO_ENDPOINT = (
    COPERNICUS_MARINE_AUTH_SYSTEM_URL
    + "realms/MIS/protocol/openid-connect/userinfo"
)

ACCEPTED_HOSTS_NETRC_FILE = [
    "nrt.cmems-du.eu",
    "my.cmems-du.eu",
    COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN,
    "default_host",
]
DEPRECATED_HOSTS = [
    "nrt.cmems-du.eu",
    "my.cmems-du.eu",
]


class CredentialsCannotBeNone(Exception):
    """
    Exception raised when credentials are not set.

    To use the Copernicus Marine Service, you need to provide a username and
    a password. You can set them as environment variables or pass them as
    arguments to the function or use the :func:`~copernicusmarine.login` command.
    To register and create your valid credentials, please visit the
    Copernicus Marine `registration page <https://data.marine.copernicus.eu/register>`_
    """

    pass


class InvalidUsernameOrPassword(Exception):
    """
    Exception raised when the username or password are invalid.

    To register and create your valid credentials, please visit:
    Copernicus Marine `registration page <https://data.marine.copernicus.eu/register>`_
    """

    pass


class CouldNotConnectToAuthenticationSystem(Exception):
    """
    Exception raised when the client could not connect to the authentication system.

    Please check the following common problems:

    - Check your internet connection
    - Make sure to authorize ``auth.marine.copernicus.eu`` domain

    If none of this worked, maybe the authentication system is down, please try again later.
    """  # noqa

    pass


def _warning_netrc_deprecated_hosts():
    logger.warning(
        "The following hosts are deprecated and will be removed in future versions: "
        f"{DEPRECATED_HOSTS}. "
        "Please update your netrc file to use the authentication system domain: "
        f"{COPERNICUS_MARINE_AUTH_SYSTEM_DOMAIN}."
    )


def _warning_motuclient_deprecated():
    logger.warning(
        "The motuclient configuration file is deprecated "
        "and will be removed in future versions. Please use "
        "the login command and or function to create a new configuration file."
    )


def _load_credential_from_copernicus_marine_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    configuration_file = open(configuration_filename)
    configuration_string = base64.standard_b64decode(
        configuration_file.read()
    ).decode("utf8")
    config = configparser.RawConfigParser()
    config.read_string(configuration_string)
    credential = config.get("credentials", credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _load_credential_from_netrc_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    authenticator = None
    for host in ACCEPTED_HOSTS_NETRC_FILE:
        authenticator = netrc(configuration_filename).authenticators(host=host)
        if (
            authenticator
            and host in DEPRECATED_HOSTS
            and credential_type == "username"
        ):
            _warning_netrc_deprecated_hosts()
        if authenticator:
            break
    if authenticator:
        username, _, password = authenticator
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
        return username if credential_type == "username" else password
    else:
        return None


def _load_credential_from_motu_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    motu_file = open(configuration_filename)
    motu_credential_type = "user" if credential_type == "username" else "pwd"
    config = configparser.RawConfigParser()
    config.read_string(motu_file.read())
    credential = config.get("Main", motu_credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _retrieve_credential_from_prompt(
    credential_type: Literal["username", "password"], hide_input: bool
) -> str:
    if credential_type == "username":
        logger.info(
            "Downloading Copernicus Marine data requires a Copernicus Marine username "
            "and password, sign up for free at:"
            " https://data.marine.copernicus.eu/register"
        )
    return click.prompt(
        "Copernicus Marine " + credential_type, hide_input=hide_input
    )


def _retrieve_credential_from_environment_variable(
    credential_type: Literal["username", "password"],
) -> Optional[str]:
    if credential_type == "username":
        logger.debug("Tried to load username from environment variable")
        return COPERNICUSMARINE_SERVICE_USERNAME
    if credential_type == "password":
        logger.debug("Tried to load password from environment variable")
        return COPERNICUSMARINE_SERVICE_PASSWORD


def _retrieve_credential_from_custom_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: pathlib.Path,
) -> Optional[str]:
    if "netrc" in str(credentials_file):
        credential = _load_credential_from_netrc_configuration_file(
            credential_type, credentials_file
        )
    elif "motuclient" in str(credentials_file):
        if credential_type == "username":
            _warning_motuclient_deprecated()
        credential = _load_credential_from_motu_configuration_file(
            credential_type, credentials_file
        )
    else:
        credential = (
            _load_credential_from_copernicus_marine_configuration_file(
                credential_type, credentials_file
            )
        )
    return credential


def _retrieve_credential_from_default_configuration_files(
    credential_type: Literal["username", "password"],
) -> Optional[str]:
    copernicus_marine_configuration_file = pathlib.Path(
        DEFAULT_CLIENT_CREDENTIALS_FILEPATH
    )
    motu_configuration_file = pathlib.Path(
        pathlib.Path.home() / "motuclient" / "motuclient-python.ini"
    )
    netrc_configuration_file = pathlib.Path(
        pathlib.Path.home() / ("_netrc" if system() == "Windows" else ".netrc")
    )
    if copernicus_marine_configuration_file.exists():
        credential = (
            _load_credential_from_copernicus_marine_configuration_file(
                credential_type,
                copernicus_marine_configuration_file,
            )
        )
    elif motu_configuration_file.exists():
        if credential_type == "username":
            _warning_motuclient_deprecated()
        credential = _load_credential_from_motu_configuration_file(
            credential_type, motu_configuration_file
        )
    elif netrc_configuration_file.exists():
        credential = _load_credential_from_netrc_configuration_file(
            credential_type, netrc_configuration_file
        )
    else:
        credential = None
    return credential


def _retrieve_credential_from_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: Optional[pathlib.Path],
) -> Optional[str]:
    if credentials_file and credentials_file.exists():
        credential = _retrieve_credential_from_custom_configuration_files(
            credential_type, credentials_file
        )
    else:
        credential = _retrieve_credential_from_default_configuration_files(
            credential_type
        )
    return credential


def copernicusmarine_validate_and_get_user(
    configuration_file: Optional[pathlib.Path],
    username: Optional[str],
    password: Optional[str],
) -> Union[str, None]:
    if username and password:
        if user := _validate_and_get_user(username, password):
            logger.info("Valid credentials from input username and password.")
            return user
        else:
            logger.error(
                "Invalid credentials from input username and password."
            )
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif (
        COPERNICUSMARINE_SERVICE_USERNAME and COPERNICUSMARINE_SERVICE_PASSWORD
    ):
        if user := _validate_and_get_user(
            COPERNICUSMARINE_SERVICE_USERNAME,
            COPERNICUSMARINE_SERVICE_PASSWORD,
        ):
            logger.info(
                "Valid credentials from environment variables: "
                "COPERNICUSMARINE_SERVICE_USERNAME and "
                "COPERNICUSMARINE_SERVICE_PASSWORD."
            )
            return user
        else:
            logger.error(
                "Invalid credentials from environment variables: "
                "COPERNICUSMARINE_SERVICE_USERNAME and "
                "COPERNICUSMARINE_SERVICE_PASSWORD."
            )
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif (
        username := _retrieve_credential_from_configuration_files(
            "username", configuration_file
        )
    ) and (
        password := _retrieve_credential_from_configuration_files(
            "password", configuration_file
        )
    ):
        if user := _validate_and_get_user(username, password):
            logger.info("Valid credentials from configuration file.")
            return user
        else:
            logger.error("Invalid credentials from configuration file.")
            logger.info(RECOVER_YOUR_CREDENTIALS_MESSAGE)
            return None
    elif configuration_file:
        logger.info(
            f"No credentials found in configuration file {configuration_file}."
        )
        logger.info(
            "Please be sure the configuration file is correct: "
            "it exists and the format is correct (especially in "
            "the case of netrc or motuclient file)."
        )
        return None
    else:
        logger.info("No credentials found.")
        logger.info(
            "Please provide credentials as arguments or environment "
            "variables, or use the 'login' command to create a credentials file."
        )
    return None


def create_copernicusmarine_configuration_file(
    username: str,
    password: str,
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
) -> tuple[Optional[pathlib.Path], bool]:
    configuration_lines = [
        "[credentials]\n",
        f"username={username}\n",
        f"password={password}\n",
    ]
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    if configuration_filename.exists() and not force_overwrite:
        confirmed = click.confirm(
            f"File {configuration_filename} already exists, overwrite it ?"
        )
        if not confirmed:
            logger.error("Abort")
            return None, True

    configuration_file_directory.mkdir(parents=True, exist_ok=True)
    configuration_file = open(configuration_filename, "w")
    configuration_string = base64.b64encode(
        "".join(configuration_lines).encode("ascii", "strict")
    ).decode("utf8")
    configuration_file.write(configuration_string)
    configuration_file.close()
    return configuration_filename, False


def _check_credentials_with_cas(
    username: str, password: str
) -> Union[str, None]:
    """
    Check the credentials with the Copernicus Marine Authentication System (CAS).
    Returns the username if the credentials are valid, None otherwise.
    Raises if the connection to the authentication system fails.
    """
    keycloak_url = COPERNICUS_MARINE_AUTH_SYSTEM_TOKEN_ENDPOINT
    client_id = "toolbox"
    scope = "openid profile email"

    data = {
        "client_id": client_id,
        "grant_type": "password",
        "username": username,
        "password": password,
        "scope": scope,
    }
    conn_session = get_configured_requests_session()
    logger.debug(f"POSTing credentials to {keycloak_url}...")
    response_post = conn_session.post(
        keycloak_url, data=data, proxies=conn_session.proxies
    )
    logger.debug(f"Response status code: {response_post.status_code}")
    if response_post.status_code == 200:
        token_response = response_post.json()
        access_token = token_response["access_token"]
        bearer_auth = BearerAuth(access_token)
        userinfo_url = COPERNICUS_MARINE_AUTH_SYSTEM_USERINFO_ENDPOINT
        logger.debug(f"GETing {userinfo_url}...")
        response_get = conn_session.get(
            userinfo_url, auth=bearer_auth, proxies=conn_session.proxies
        )
        response_get.raise_for_status()
        if response_get.status_code == 200:
            response_json = response_get.json()
            return response_json["preferred_username"]
    elif response_post.status_code == 401:
        # Invalid credentials
        return None
    else:
        response_post.raise_for_status()

    conn_session.close()
    return None


def _validate_and_get_user(username: str, password: str) -> Union[str, None]:
    try:
        result = _get_user(username, password)
        return result
    except Exception as e:
        logger.debug(
            f"Could not connect with the authentication system because of: {e}"
        )
        raise CouldNotConnectToAuthenticationSystem()


def _get_user(username: str, password: str) -> Union[str, None]:
    number_of_retry = 3
    while number_of_retry > 0:
        try:
            user = _check_credentials_with_cas(
                username=username, password=password
            )
            return user
        except requests.exceptions.ConnectTimeout:
            number_of_retry -= 1
        except requests.exceptions.ConnectionError:
            number_of_retry -= 1
    raise CouldNotConnectToAuthenticationSystem()


def get_credential(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
    credentials_file: Optional[pathlib.Path],
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_configuration_files(
                credential_type=credential_type,
                credentials_file=credentials_file,
            )
            if not credential:
                credential = _retrieve_credential_from_prompt(
                    credential_type, hide_input=hide_input
                )
                if not credential:
                    raise ValueError(f"{credential} cannot be None")
    else:
        logger.debug("Credentials loaded from function arguments")
    return credential


def get_and_check_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
) -> tuple[str, str]:
    username, password = get_username_password(
        username=username, password=password, credentials_file=credentials_file
    )
    user = _validate_and_get_user(
        username,
        password,
    )
    if not user:
        raise InvalidUsernameOrPassword(
            "Learn how to recover your credentials at: "
            "https://help.marine.copernicus.eu/en/articles/"
            "4444552-i-forgot-my-username-or-my-password-what-should-i-do"
        )
    return (user, password)


def get_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
) -> tuple[str, str]:
    username = get_credential(
        username,
        "username",
        hide_input=False,
        credentials_file=credentials_file,
    )
    password = get_credential(
        password,
        "password",
        hide_input=True,
        credentials_file=credentials_file,
    )
    return (username, password)


def _get_credential_from_environment_variable_or_prompt(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_prompt(
                credential_type, hide_input
            )
            if not credential:
                raise CredentialsCannotBeNone(credential_type)
    return credential


def credentials_file_builder(
    username: Optional[str],
    password: Optional[str],
    configuration_file_directory: pathlib.Path,
    force_overwrite: bool,
) -> tuple[Optional[pathlib.Path], bool]:
    """
    Returns:
        a path to the configuration file: if none, the credentials are not valid

        a boolean that indicates
        if the user aborted the creation of the configuration file
    """
    username = _get_credential_from_environment_variable_or_prompt(
        username, "username", False
    )
    password = _get_credential_from_environment_variable_or_prompt(
        password, "password", True
    )
    copernicus_marine_credentials_are_valid = _validate_and_get_user(
        username, password
    )
    if copernicus_marine_credentials_are_valid:
        (
            configuration_file,
            has_been_aborted,
        ) = create_copernicusmarine_configuration_file(
            username=username,
            password=password,
            configuration_file_directory=configuration_file_directory,
            force_overwrite=force_overwrite,
        )
        if has_been_aborted:
            return configuration_file, has_been_aborted
        if configuration_file:
            return configuration_file, False
    return None, False



import json
import logging
import pathlib
from typing import Optional

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.exceptions import (
    ServiceNotSupported,
    WrongFormatRequested,
)
from copernicusmarine.core_functions.marine_datastore_config import (
    get_config_and_check_version_subset,
)
from copernicusmarine.core_functions.models import CommandType, ResponseSubset
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.services_utils import (
    RetrievalService,
    get_retrieval_service,
)
from copernicusmarine.core_functions.utils import get_unique_filepath
from copernicusmarine.download_functions.download_sparse import download_sparse
from copernicusmarine.download_functions.download_zarr import download_zarr
from copernicusmarine.download_functions.subset_xarray import (
    check_dataset_subset_bounds,
)

logger = logging.getLogger("copernicusmarine")


def subset_function(
    subset_request: SubsetRequest,
) -> ResponseSubset:
    retrieval_service = retrieve_metadata_and_check_request(subset_request)
    if retrieval_service.service_name not in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.PLATFORMSERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        raise ServiceNotSupported(retrieval_service.service_name)
    return download_zarr_or_sparse(
        subset_request=subset_request,
        retrieval_service=retrieval_service,
        tdqm_configuration={"disable": subset_request.disable_progress_bar},
    )


def retrieve_metadata_and_check_request(
    subset_request: SubsetRequest,
) -> RetrievalService:
    marine_datastore_config = get_config_and_check_version_subset(
        subset_request.staging
    )

    retrieval_service: RetrievalService = get_retrieval_service(
        request=subset_request,
        command_type=CommandType.SUBSET,
        marine_datastore_config=marine_datastore_config,
    )

    check_requested_area_time_valid(
        subset_request=subset_request,
        service_format=retrieval_service.service_format,
        dataset_part=retrieval_service.dataset_part.name,
    )
    check_dataset_subset_bounds(
        service=retrieval_service.service,
        part=retrieval_service.dataset_part,
        dataset_subset=subset_request,
        coordinates_selection_method=subset_request.coordinates_selection_method,
        axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
    )
    return retrieval_service


def download_zarr_or_sparse(
    subset_request: SubsetRequest,
    retrieval_service: RetrievalService,
    tdqm_configuration: dict,
) -> ResponseSubset:
    if retrieval_service.service_format == CopernicusMarineServiceFormat.ZARR:
        raise_when_all_dataset_requested(subset_request, False)
        if "file_format" not in subset_request.model_fields_set:
            subset_request.file_format = "netcdf"
        elif subset_request.file_format not in ["netcdf", "zarr"]:
            raise WrongFormatRequested(
                requested_format=subset_request.file_format,
                supported_formats=["netcdf", "zarr"],
            )
        logger.debug(
            f"Downloading data in {subset_request.file_format} format."
        )
        return download_zarr(
            subset_request=subset_request,
            dataset_url=retrieval_service.uri,
            dataset_valid_start_date=retrieval_service.dataset_valid_start_date,
            service=retrieval_service.service,
            is_original_grid=retrieval_service.is_original_grid,
            axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
            dataset_chunking=retrieval_service.dataset_chunking,
            tdqm_configuration=tdqm_configuration,
        )
    if (
        retrieval_service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ):
        raise_when_all_dataset_requested(subset_request, True)
        if "file_format" not in subset_request.model_fields_set:
            subset_request.file_format = "csv"
        elif subset_request.file_format not in ["parquet", "csv"]:
            raise WrongFormatRequested(
                requested_format=subset_request.file_format,
                supported_formats=["parquet", "csv"],
            )
        logger.debug(
            f"Downloading data in {subset_request.file_format} format."
        )
        if subset_request.coordinates_selection_method not in [
            "inside",
            "strict-inside",
        ]:
            logger.warning(
                f"coordinates-selection-method "
                f"{subset_request.coordinates_selection_method} "
                "is not supported for sparse data. "
                "Using 'inside' by default."
            )
        return download_sparse(
            subset_request.username,
            subset_request,
            retrieval_service.metadata_url,
            retrieval_service.service,
            retrieval_service.axis_coordinate_id_mapping,
            retrieval_service.product_doi,
            subset_request.disable_progress_bar,
        )
    else:
        raise ValueError("Dataset is missing some metadata. Cannot subset")


def create_subset_template() -> None:
    filename = pathlib.Path("subset_template.json")
    if filename.exists():
        get_unique_filepath(
            filepath=filename,
        )
    with open(filename, "w") as output_file:
        json.dump(
            {
                "dataset_id": "cmems_mod_ibi_phy-temp_my_0.027deg_P1M-m",
                "start_datetime": "2023-10-01",
                "end_datetime": "2023-11-01",
                "minimum_longitude": -85,
                "maximum_longitude": -10,
                "minimum_latitude": 35,
                "maximum_latitude": 43,
                "minimum_depth": 1,
                "maximum_depth": 10,
                "variables": ["bottomT", "thetao"],
                "output_directory": "copernicusmarine_data",
                "service": None,
                "overwrite": False,
                "dry_run": False,
            },
            output_file,
            indent=4,
        )
    logger.info(f"Template created at: {filename}")


def raise_when_all_dataset_requested(
    subset_request: SubsetRequest, sparse_data: bool
) -> None:
    """
    Raise an error if the subset request is not a subsetting request
    and all the dataset would be downloaded.

    Parameters
    ----------
    subset_request: SubsetRequest
        The subset request.
    sparse_data: bool
        If the requested dataset is sparse data.
        If True, only subsetting on platform_ids is allowed.
        Otherwise it should raise.
    """
    if all(
        e is None
        for e in [
            subset_request.variables,
            subset_request.minimum_x,
            subset_request.maximum_x,
            subset_request.minimum_y,
            subset_request.maximum_y,
            subset_request.minimum_depth,
            subset_request.maximum_depth,
            subset_request.start_datetime,
            subset_request.end_datetime,
            (sparse_data and subset_request.platform_ids) or None,
        ]
    ):
        logger.info(
            "To retrieve a complete dataset, please use instead: "
            f"copernicusmarine get --dataset-id {subset_request.dataset_id}"
        )
        raise ValueError(
            "Missing subset option. Try 'copernicusmarine subset --help'."
        )


def check_requested_area_time_valid(
    subset_request: SubsetRequest,
    service_format: Optional[CopernicusMarineServiceFormat],
    dataset_part: str,
) -> None:
    is_original_grid = dataset_part == "originalGrid"
    x_axis_name = "x" if is_original_grid else "longitude"
    y_axis_name = "y" if is_original_grid else "latitude"
    if (
        subset_request.minimum_x is not None
        and subset_request.maximum_x is not None
        and subset_request.minimum_x > subset_request.maximum_x
    ):
        if (
            service_format == CopernicusMarineServiceFormat.ZARR
            and not is_original_grid
        ):
            logger.warning(
                "Minimum longitude greater than maximum longitude. "
                "Your selection will wrap around the 180° meridian: "
                f"Minimum longitude: {subset_request.minimum_x}, "
                f"Maximum longitude: {subset_request.maximum_x}."
            )
        else:
            raise ValueError(
                f"Minimum {x_axis_name} greater than maximum {x_axis_name}: "
                f"minimum-{x_axis_name} option must be smaller or equal to "
                f"maximum-{x_axis_name} for this dataset. "
                f"Minimum {x_axis_name}: {subset_request.minimum_x}, "
                f"Maximum {x_axis_name}: {subset_request.maximum_x}."
            )
    if (
        subset_request.minimum_y is not None
        and subset_request.maximum_y is not None
        and subset_request.minimum_y > subset_request.maximum_y
    ):
        raise ValueError(
            f"Minimum {y_axis_name} greater than maximum {y_axis_name}: "
            f"minimum-{y_axis_name} option must be smaller or equal to "
            f"maximum-{y_axis_name} for this dataset. "
            f"Minimum {y_axis_name}: {subset_request.minimum_y}, "
            f"Maximum {y_axis_name}: {subset_request.maximum_y}."
        )
    if (
        subset_request.minimum_depth is not None
        and subset_request.maximum_depth is not None
        and subset_request.minimum_depth > subset_request.maximum_depth
    ):
        raise ValueError(
            "Minimum depth greater than maximum depth: minimum-depth "
            "option must be smaller or equal to maximum-depth. "
            f"Minimum depth: {subset_request.minimum_depth}, "
            f"Maximum depth: {subset_request.maximum_depth}."
        )
    if (
        subset_request.start_datetime is not None
        and subset_request.end_datetime is not None
        and subset_request.start_datetime > subset_request.end_datetime
    ):
        raise ValueError(
            "Start datetime greater than end datetime: start-datetime "
            "option must be smaller or equal to end-datetime. "
            f"Start date: {subset_request.start_datetime}, "
            f"End date: {subset_request.end_datetime}."
        )



import logging
import os
import pathlib
import warnings
from copy import deepcopy
from multiprocessing import current_process
from typing import Optional, Tuple, Union

import xarray
import zarr

from copernicusmarine.core_functions.temporary_path_saver import (
    TemporaryPathSaver,
)

if zarr.__version__.startswith("2"):
    from zarr.storage import DirectoryStore

    # If zarr client is version 2 there are no
    # `zarr_format` argument
    ZARR_FORMAT = None
else:
    from zarr.storage import LocalStore as DirectoryStore

    # for client zarr>2 we need to specify the format
    # otherwise it uses zarr format 3 by default
    ZARR_FORMAT = 2

from tqdm.dask import TqdmCallback

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineCoordinate,
    CopernicusMarineService,
)
from copernicusmarine.core_functions import custom_open_zarr
from copernicusmarine.core_functions.exceptions import (
    NetCDFCompressionNotAvailable,
)
from copernicusmarine.core_functions.models import (
    CoordinatesSelectionMethod,
    DatasetChunking,
    FileStatus,
    ResponseSubset,
    StatusCode,
    StatusMessage,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.utils import (
    add_copernicusmarine_version_in_dataset_attributes,
    get_unique_filepath,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    TemporalParameters,
)
from copernicusmarine.download_functions.subset_xarray import subset
from copernicusmarine.download_functions.utils import (
    get_approximation_size_data_downloaded,
    get_approximation_size_final_result,
    get_dataset_coordinates_extent,
    get_filename,
    timestamp_or_datestring_to_datetime,
)

logger = logging.getLogger("copernicusmarine")


def get_dataset_and_parameters(
    subset_request: SubsetRequest,
    dataset_url: str,
    axis_coordinate_id_mapping: dict[str, str],
    service: CopernicusMarineService,
    dataset_chunking: Optional[DatasetChunking],
    is_original_grid: bool,
    dataset_valid_start_date: Optional[Union[str, int, float]],
) -> tuple[xarray.Dataset, GeographicalParameters, DepthParameters]:
    if dataset_valid_start_date:
        minimum_start_date = timestamp_or_datestring_to_datetime(
            dataset_valid_start_date
        )
        if (
            not subset_request.start_datetime
            or subset_request.start_datetime < minimum_start_date
        ):
            subset_request.start_datetime = minimum_start_date

    if subset_request.chunk_size_limit and dataset_chunking:
        optimum_dask_chunking = get_optimum_dask_chunking(
            service=service,
            variables=subset_request.variables,
            dataset_chunking=dataset_chunking,
            chunk_size_limit=subset_request.chunk_size_limit,
            axis_coordinate_id_mapping=axis_coordinate_id_mapping,
        )
    else:
        optimum_dask_chunking = None

    logger.debug(f"Dask chunking selected: {optimum_dask_chunking}")
    geographical_parameters = subset_request.get_geographical_parameters(
        axis_coordinate_id_mapping, is_original_grid
    )
    depth_parameters = subset_request.get_depth_parameters(
        axis_coordinate_id_mapping,
    )
    dataset = open_dataset_from_arco_series(
        username=subset_request.username,
        dataset_url=dataset_url,
        variables=subset_request.variables,
        geographical_parameters=geographical_parameters,
        temporal_parameters=subset_request.get_temporal_parameters(
            axis_coordinate_id_mapping,
        ),
        depth_parameters=depth_parameters,
        coordinates_selection_method=subset_request.coordinates_selection_method,
        optimum_dask_chunking=optimum_dask_chunking,
    )

    dataset = add_copernicusmarine_version_in_dataset_attributes(dataset)
    return dataset, geographical_parameters, depth_parameters


def download_zarr(
    subset_request: SubsetRequest,
    dataset_url: str,
    dataset_valid_start_date: Optional[Union[str, int, float]],
    service: CopernicusMarineService,
    axis_coordinate_id_mapping: dict[str, str],
    dataset_chunking: Optional[DatasetChunking],
    is_original_grid: bool,
    tdqm_configuration: dict,
) -> ResponseSubset:
    (
        dataset,
        geographical_parameters,
        depth_parameters,
    ) = get_dataset_and_parameters(
        subset_request=subset_request,
        dataset_url=dataset_url,
        axis_coordinate_id_mapping=axis_coordinate_id_mapping,
        service=service,
        dataset_chunking=dataset_chunking,
        is_original_grid=is_original_grid,
        dataset_valid_start_date=dataset_valid_start_date,
    )
    if depth_parameters.vertical_axis == "elevation":
        axis_coordinate_id_mapping["z"] = "elevation"

    if not subset_request.output_directory.is_dir():
        pathlib.Path.mkdir(subset_request.output_directory, parents=True)

    logger.debug(f"Xarray Dataset: {dataset}")
    logger.debug("Starting download. Please wait...")

    total_size_estimation = get_approximation_size_final_result(
        dataset, axis_coordinate_id_mapping
    )

    logger.debug(f"Total size estimation: {total_size_estimation} MB")

    dataset.close()

    filename = get_filename(
        subset_request.output_filename,
        dataset,
        subset_request.dataset_id,
        subset_request.file_format,
        axis_coordinate_id_mapping,
        geographical_parameters,
    )

    final_result_size_estimation = get_approximation_size_final_result(
        dataset, axis_coordinate_id_mapping
    )
    if dataset_chunking:
        data_needed_approximation = get_approximation_size_data_downloaded(
            dataset, dataset_chunking
        )
    else:
        data_needed_approximation = None

    output_path = pathlib.Path(subset_request.output_directory, filename)

    if not subset_request.overwrite and not subset_request.skip_existing:
        output_path = get_unique_filepath(
            filepath=output_path,
        )

    response = ResponseSubset(
        file_path=output_path,
        output_directory=subset_request.output_directory,
        filename=output_path.name,
        file_size=final_result_size_estimation,
        data_transfer_size=data_needed_approximation,
        variables=list(dataset.data_vars),
        coordinates_extent=get_dataset_coordinates_extent(
            dataset, axis_coordinate_id_mapping
        ),
        status=StatusCode.SUCCESS,
        message=StatusMessage.SUCCESS,
        file_status=FileStatus.DOWNLOADED,
    )

    if subset_request.dry_run:
        response.status = StatusCode.DRY_RUN
        response.message = StatusMessage.DRY_RUN
        return response
    if subset_request.skip_existing and os.path.exists(output_path):
        response.file_status = FileStatus.IGNORED
        return response

    current = current_process()
    if "position" not in tdqm_configuration and current._identity:
        tdqm_configuration["position"] = current._identity[0]
    with TqdmCallback(
        **tdqm_configuration,
    ):
        _save_dataset_locally(
            dataset,
            output_path,
            subset_request.netcdf_compression_level,
            subset_request.netcdf3_compatible,
        )

    dataset.close()

    if subset_request.overwrite:
        response.status = StatusCode.SUCCESS
        response.message = StatusMessage.SUCCESS
        response.file_status = FileStatus.OVERWRITTEN

    return response


def open_dataset_from_arco_series(
    username: str,
    dataset_url: str,
    variables: Optional[list[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
    coordinates_selection_method: CoordinatesSelectionMethod,
    optimum_dask_chunking: Optional[dict[str, int]],
) -> xarray.Dataset:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        dataset = custom_open_zarr.open_zarr(
            dataset_url,
            chunks=optimum_dask_chunking,
            copernicus_marine_username=username,
        )
    for variable in dataset:
        del dataset[variable].encoding["chunks"]
    dataset = subset(
        dataset=dataset,
        variables=variables,
        geographical_parameters=geographical_parameters,
        temporal_parameters=temporal_parameters,
        depth_parameters=depth_parameters,
        coordinates_selection_method=coordinates_selection_method,
    )
    if "depth" in dataset.coords and optimum_dask_chunking:
        optimum_chunks_depth = deepcopy(optimum_dask_chunking)
        if "elevation" in optimum_chunks_depth:
            optimum_chunks_depth["depth"] = optimum_chunks_depth["elevation"]
            del optimum_chunks_depth["elevation"]
        dataset = dataset.chunk(optimum_chunks_depth)
    elif optimum_dask_chunking:
        dataset = dataset.chunk(optimum_dask_chunking)
    return dataset


def get_coordinates_dask_and_zarr_chunks_info(
    service: CopernicusMarineService,
    variables: Optional[list[str]],
    dataset_chunking: DatasetChunking,
) -> Tuple[dict, dict]:
    """
    Return
    -------

    Tuple[dict, dict]
        A tuple containing:
        - a dict with the maximum dask chunk factor for each coordinate id.
          It tells us how many times we can multiply the zarr chunking per coordinate.
        - a dict with the zarr chunking for each coordinate id. i.e. the number of values per zarr chunk.
    """  # noqa
    set_variables = set(variables) if variables else set()
    coordinate_for_subset: dict[str, CopernicusMarineCoordinate] = {}
    for variable in service.variables:
        if (
            not set_variables
            or variable.short_name in set_variables
            or variable.standard_name in set_variables
        ):
            for coordinate in variable.coordinates:
                coordinate_name = coordinate.coordinate_id
                if coordinate.coordinate_id not in coordinate_for_subset:
                    coordinate_for_subset[coordinate_name] = coordinate

    max_dask_chunk_factor = {}
    coordinate_zarr_chunk_length = {}
    for coordinate_id, coordinate in coordinate_for_subset.items():
        chunking_length = coordinate.chunking_length
        if chunking_length is None:
            continue
        coordinate_zarr_chunk_length[coordinate_id] = int(chunking_length)
        number_of_zarr_chunks_needed = (
            dataset_chunking.get_number_chunks_coordinate(coordinate_id)
        )
        if number_of_zarr_chunks_needed is None:
            continue
        max_dask_chunk_factor[coordinate_id] = number_of_zarr_chunks_needed
    return (
        max_dask_chunk_factor,
        coordinate_zarr_chunk_length,
    )


def get_optimum_dask_chunking(
    service: CopernicusMarineService,
    variables: Optional[list[str]],
    dataset_chunking: DatasetChunking,
    chunk_size_limit: int,
    axis_coordinate_id_mapping: dict[str, str],
) -> Optional[dict[str, int]]:
    """
    We have some problems with overly big dask graphs (we think) that introduces huge overheads
    and memory usage. We are trying to find the optimum chunking for dask arrays.

    By default, the chunking is the zarr chunking ie 1MB for each tile.
    The rule of thumb for dask chunking is to have chunks of size 100MB.

    To avoid downloading too much data we should also be careful not to
    increase the size of the chunks more than the size of the data we are interested in.
    Eg: you want 1 time point, increasing size of the chunk on time dimension by x
    will lead to downloading x times more data than needed.

    Knowing that, we should cap the size of the chunk to 100MB and use multiples of the zarr chunking.

    If the factors sum up to less than 100, no chunking is needed.

    Returns
    -------

    Optional[dict[str, int]]
        A dictionary with the optimum dask chunking for each coordinate id. In the form:
        {
            "longitude": 2,
            "latitude": 2,
            "time": 1,
            "elevation": 1, # forced to work with ARCO standard
        }
        Can also be None if the dataset is not large enough to require dask chunking.
    """  # noqa
    (
        max_dask_chunk_factor,
        coordinate_zarr_chunk_length,
    ) = get_coordinates_dask_and_zarr_chunks_info(
        service, variables, dataset_chunking
    )
    zarr_chunks_to_download = dataset_chunking.number_chunks
    # it seems that dask chunks increases more
    # with the number of variables
    chunks_and_variables_size = zarr_chunks_to_download * len(
        dataset_chunking.chunking_per_variable
    )
    if chunk_size_limit == -1:
        # TODO: investigate in depth what are the optimal values.
        # For now, we use the default values that seem to work well
        # for our examples but we should investigate more.
        if zarr_chunks_to_download <= 50:
            return None
        elif chunks_and_variables_size <= 1500:
            chunk_size_limit = 20
        elif chunks_and_variables_size <= 4000:
            chunk_size_limit = 50
        else:
            chunk_size_limit = 100
    logger.debug(f"Chunk size limit: {chunk_size_limit}")
    logger.debug(f"Zarr chunking: {coordinate_zarr_chunk_length}")
    logger.debug(f"Max dask chunk factor: {max_dask_chunk_factor}")
    optimum_dask_factors = _get_optimum_factors(
        max_dask_chunk_factor,
        chunk_size_limit,
        axis_coordinate_id_mapping,
    )
    optimum_dask_chunking = {
        coordinate_id: coordinate_zarr_chunk_length[coordinate_id]
        * optimum_dask_factors[coordinate_id]
        for coordinate_id in optimum_dask_factors
    }
    logger.debug(f"Optimum dask chunking: {optimum_dask_chunking}")

    if (
        "z" in axis_coordinate_id_mapping
        and (z_coordinate := axis_coordinate_id_mapping["z"]) != "elevation"
        and z_coordinate in optimum_dask_chunking
    ):
        optimum_dask_chunking["elevation"] = optimum_dask_chunking[
            z_coordinate
        ]
        del optimum_dask_chunking[z_coordinate]
    return optimum_dask_chunking


def _product(iterable) -> int:
    result = 1
    for i in iterable:
        result *= i
    return result


def _get_optimum_factors(
    coordinate_max_dask_chunk_factor: dict[str, int],
    limit: int,
    axis_coordinate_id_mapping: dict[str, str],
) -> dict[str, int]:
    optimum_factors = {
        coordinate_name: 1
        for coordinate_name in coordinate_max_dask_chunk_factor
    }
    coordinate_selection_pool = {
        coordinate_name
        for coordinate_name, max_factor in coordinate_max_dask_chunk_factor.items()
        if max_factor > 1
    }
    while (
        _product(optimum_factors.values()) < limit
        and coordinate_selection_pool
    ):
        selected_coordinate = max(
            coordinate_selection_pool,
            key=lambda x: (
                coordinate_max_dask_chunk_factor[x],
                x in axis_coordinate_id_mapping.get("z", ""),
                x in axis_coordinate_id_mapping.get("t", ""),
                x in axis_coordinate_id_mapping.get("y", ""),
                x in axis_coordinate_id_mapping.get("x", ""),
            ),
        )

        optimum_factors[selected_coordinate] += 1
        if optimum_factors[
            selected_coordinate
        ] == coordinate_max_dask_chunk_factor[selected_coordinate] or (
            len(coordinate_selection_pool) > 1
            and optimum_factors[selected_coordinate] >= limit // 2
        ):
            coordinate_selection_pool.remove(selected_coordinate)
    return optimum_factors


def _save_dataset_locally(
    dataset: xarray.Dataset,
    output_path: pathlib.Path,
    netcdf_compression_level: int,
    netcdf3_compatible: bool,
):
    with TemporaryPathSaver(output_path) as temp_path:
        if output_path.suffix == ".zarr":
            if netcdf_compression_level > 0:
                raise NetCDFCompressionNotAvailable(
                    "--netcdf-compression-level option cannot be used when "
                    "writing to ZARR"
                )
            _download_dataset_as_zarr(dataset, temp_path)
        else:
            _download_dataset_as_netcdf(
                dataset,
                temp_path,
                netcdf_compression_level,
                netcdf3_compatible,
            )


def _download_dataset_as_zarr(
    dataset: xarray.Dataset, output_path: pathlib.Path
):
    logger.debug("Writing dataset to Zarr")
    store = DirectoryStore(output_path)
    if ZARR_FORMAT is None:
        return dataset.to_zarr(store=store, mode="w")
    else:
        return dataset.to_zarr(store=store, mode="w", zarr_format=ZARR_FORMAT)


def _download_dataset_as_netcdf(
    dataset: xarray.Dataset,
    output_path: pathlib.Path,
    netcdf_compression_level: int,
    netcdf3_compatible: bool,
):
    logger.debug("Writing dataset to NetCDF")
    for coord in dataset.coords:
        dataset[coord].encoding["_FillValue"] = None
    if netcdf_compression_level > 0:
        logger.info(
            f"NetCDF compression enabled with level {netcdf_compression_level}"
        )
        comp = {
            "zlib": True,
            "complevel": netcdf_compression_level,
            "contiguous": False,
            "shuffle": True,
        }

        keys_to_keep = {
            "scale_factor",
            "add_offset",
            "dtype",
            "_FillValue",
            "units",
        }
        encoding = {
            name: {
                **{
                    key: value
                    for key, value in var.encoding.items()
                    if key in keys_to_keep
                },
                **comp,
            }
            for name, var in dataset.data_vars.items()
        }
    else:
        encoding = None

    xarray_download_format = "NETCDF3_CLASSIC" if netcdf3_compatible else None
    engine = "h5netcdf" if not netcdf3_compatible else "netcdf4"
    return dataset.to_netcdf(
        output_path,
        mode="w",
        encoding=encoding,
        format=xarray_download_format,
        engine=engine,
    )



import pathlib
from dataclasses import dataclass
from enum import Enum
from typing import Literal, Optional, Union, get_args

from pydantic import BaseModel, ConfigDict

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineServiceNames,
    short_name_from_service_name,
)

FileFormat = Literal["netcdf", "zarr", "csv", "parquet"]
DEFAULT_FILE_FORMAT: FileFormat = "netcdf"
DEFAULT_FILE_FORMATS = list(get_args(FileFormat))

FileExtension = Literal[".nc", ".zarr"]
DEFAULT_FILE_EXTENSION: FileExtension = ".nc"
DEFAULT_FILE_EXTENSIONS = list(get_args(FileExtension))

CoordinatesSelectionMethod = Literal[
    "inside", "strict-inside", "nearest", "outside"
]
DEFAULT_COORDINATES_SELECTION_METHOD: CoordinatesSelectionMethod = "inside"
DEFAULT_COORDINATES_SELECTION_METHODS = list(
    get_args(CoordinatesSelectionMethod)
)

VerticalAxis = Literal["depth", "elevation"]
DEFAULT_VERTICAL_AXIS: VerticalAxis = "depth"
DEFAULT_VERTICAL_AXES = list(get_args(VerticalAxis))

GeoSpatialProjection = Literal["lonlat", "originalGrid"]
DEFAULT_GEOSPATIAL_PROJECTION: GeoSpatialProjection = "lonlat"

SplitOnTimeOption = Literal["hour", "day", "month", "year"]
DEFAULT_SPLIT_ON_TIME_OPTIONS = list(get_args(SplitOnTimeOption))


class ChunkType(str, Enum):
    ARITHMETIC = "default"
    GEOMETRIC = "symmetricGeometric"


class StatusCode(str, Enum):
    """
    Enumeration of the possible of a request.
    Only concerns ``get`` and ``subset`` functions.
    """

    SUCCESS = "000"
    DRY_RUN = "001"
    FILE_LIST_CREATED = "002"
    NO_DATA_TO_DOWNLOAD = "003"
    ERROR = "100"


class StatusMessage(str, Enum):
    """
    Enumeration of the possible messages of a request.
    Only concerns ``get`` and ``subset`` functions.
    """

    SUCCESS = "The request was successful."
    DRY_RUN = "The request was run with the dry-run option. No data was downloaded."  # noqa: E501
    FILE_LIST_CREATED = "The request created a file list and then stopped."
    NO_DATA_TO_DOWNLOAD = "No data to download from the remote server corresponding to your request."  # noqa: E501
    ERROR = "An error occurred during the request."


class FileStatus(str, Enum):
    """
    Gives an indication of how the file was handled.

    This depends on the option passed to the query.
    """

    #: The file has been downloaded successfully.
    DOWNLOADED = "DOWNLOADED"
    #: The file has been ignored. Probably because it already exists.
    IGNORED = "IGNORED"
    #: The file has been overwritten and downloaded.
    OVERWRITTEN = "OVERWRITTEN"

    @classmethod
    def get_status(cls, ignore: bool, overwrite: bool) -> "FileStatus":
        if ignore:
            return FileStatus.IGNORED
        if overwrite:
            return FileStatus.OVERWRITTEN
        return FileStatus.DOWNLOADED


class FileGet(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    #: Full url of the location of the file on remote server.
    s3_url: str
    #: Location of the file on the remote server, https url.
    https_url: str
    #: Size of the file in MB.
    file_size: float
    #: Last modified date.
    last_modified_datetime: str
    #: ETag of the file.
    etag: str
    #: File format.
    file_format: str
    #: Output directory where the file is stored.
    output_directory: pathlib.Path
    #: File name.
    filename: str
    #: Relative path to the file.
    file_path: pathlib.Path
    #: Status of the file.
    file_status: FileStatus


class ResponseGet(BaseModel):
    """Metadata returned when using :func:`~copernicusmarine.get`"""

    model_config = ConfigDict(use_enum_values=True)

    #: Description of the files concerned by the query.
    files: list[FileGet]
    #: List of deleted files. Only if option ``sync-delete`` is passed.
    files_deleted: Optional[list[str]]
    #: List of not found files from the file list input.
    files_not_found: Optional[list[str]]
    #: Number of files to be downloaded.
    number_of_files_to_download: int
    #: Total size of the files that would be downloaded in MB.
    total_size: Optional[float]
    #: Status of the request.
    status: StatusCode
    #: Message explaning the status.
    message: StatusMessage


class S3FileInfo(BaseModel):
    """
    Basic information about a file in the bucket.
    """

    model_config = ConfigDict(use_enum_values=True)

    filename_in: str
    filename_out: pathlib.Path = pathlib.Path()
    size: float
    last_modified: str
    etag: str
    ignore: bool = False
    overwrite: bool = False


class S3FilesDescriptor(BaseModel):
    model_config = ConfigDict(use_enum_values=True)

    endpoint: str = ""
    bucket: str = ""
    s3_files: list[S3FileInfo] = []
    total_size: float = 0.0
    files_not_found: list[str] = []
    files_to_delete: list[pathlib.Path] = []
    create_file_list: bool = False

    def add_s3_file(self, s3_file: S3FileInfo):
        self.s3_files.append(s3_file)
        if not s3_file.ignore:
            self.total_size += s3_file.size


class GeographicalExtent(BaseModel):
    """Interval for geographical coordinates."""

    minimum: float
    maximum: float
    unit: Optional[str]
    coordinate_id: str


class TimeExtent(BaseModel):
    """Interval for time coordinates."""

    minimum: str
    maximum: str
    unit: str
    coordinate_id: str


class ResponseSubset(BaseModel):
    """Metadata returned when using :func:`~copernicusmarine.subset`"""

    model_config = ConfigDict(use_enum_values=True)

    #: Relative path to the file.
    file_path: pathlib.Path
    #: Output directory where the file stored.
    output_directory: pathlib.Path
    #: File name.
    filename: str
    #: Estimation of the size of the final result file in MB.
    #: This estimation may not be accurate if you save the result as
    #: a compressed NetCDF file.
    file_size: Optional[float]
    #: Estimation of the maximum amount of data needed to
    #: get the final result in MB.
    data_transfer_size: Optional[float]
    #: Variables of the subsetted dataset.
    variables: list[str]
    #: The bounds of the subsetted dataset.
    coordinates_extent: list[Union[GeographicalExtent, TimeExtent]]
    #: Status of the request.
    status: StatusCode
    #: Message explaning the status.
    message: StatusMessage
    #: Status of the files.
    file_status: FileStatus


# Internal use only
@dataclass
class VariableChunking:
    variable_short_name: str
    number_values: float
    number_chunks: int
    chunk_size: float


@dataclass
class CoordinateChunking:
    coordinate_id: str
    chunking_length: float
    number_of_chunks: int


@dataclass
class DatasetChunking:
    number_chunks: int
    chunking_per_variable: dict[str, VariableChunking]
    chunking_per_coordinate: dict[str, CoordinateChunking]

    def get_number_values_variable(self, variable_short_name: str) -> float:
        if variable_short_name in self.chunking_per_variable:
            return self.chunking_per_variable[
                variable_short_name
            ].number_values
        return 0

    def get_number_chunks_coordinate(
        self, coordinate_id: str
    ) -> Optional[float]:
        if coordinate_id in self.chunking_per_coordinate:
            return self.chunking_per_coordinate[coordinate_id].number_of_chunks
        return None


class _Command(Enum):
    GET = "get"
    SUBSET = "subset"
    OPEN_DATASET = "open_dataset"
    READ_DATAFRAME = "read_dataframe"


@dataclass(frozen=True)
class Command:
    command_name: _Command
    service_names_by_priority: list[CopernicusMarineServiceNames]

    def service_names(self) -> list[str]:
        return [
            service_name.value
            for service_name in self.service_names_by_priority
        ]

    def short_names_services(self) -> list[str]:
        return [
            short_name_from_service_name(service_name).value
            for service_name in self.service_names_by_priority
        ]

    def get_available_service_for_command(self) -> list[str]:
        available_services = []
        for service_name in self.service_names_by_priority:
            available_services.append(service_name.value)
            short_name = short_name_from_service_name(service_name)
            if short_name != service_name:
                available_services.append(
                    short_name_from_service_name(service_name).value
                )
        return available_services


class CommandType(Command, Enum):
    SUBSET = (
        _Command.SUBSET,
        [
            CopernicusMarineServiceNames.GEOSERIES,
            CopernicusMarineServiceNames.TIMESERIES,
            CopernicusMarineServiceNames.OMI_ARCO,
            CopernicusMarineServiceNames.STATIC_ARCO,
            CopernicusMarineServiceNames.PLATFORMSERIES,
        ],
    )
    GET = (
        _Command.GET,
        [
            CopernicusMarineServiceNames.FILES,
        ],
    )
    OPEN_DATASET = (
        _Command.OPEN_DATASET,
        [
            CopernicusMarineServiceNames.GEOSERIES,
            CopernicusMarineServiceNames.TIMESERIES,
            CopernicusMarineServiceNames.OMI_ARCO,
            CopernicusMarineServiceNames.STATIC_ARCO,
        ],
    )
    READ_DATAFRAME = (
        _Command.READ_DATAFRAME,
        [
            CopernicusMarineServiceNames.GEOSERIES,
            CopernicusMarineServiceNames.TIMESERIES,
            CopernicusMarineServiceNames.OMI_ARCO,
            CopernicusMarineServiceNames.STATIC_ARCO,
            CopernicusMarineServiceNames.PLATFORMSERIES,
        ],
    )



import os
import pathlib
import shutil
import tempfile
import uuid
from typing import Optional


class TemporaryPathSaver:
    """
    It creates a temporary file (or directory, for Zarr) in the same folder
    as the target path, and renames it to the final destination only if
    the operation completes successfully.
    """

    def __init__(self, output_path: pathlib.Path):
        self.output_path = output_path.resolve()
        self.dir_path = self.output_path.parent
        self.base_name = self.output_path.name
        self.is_zarr = self.output_path.suffix == ".zarr"

        self.temp_path: Optional[pathlib.Path] = None

    def __enter__(self) -> pathlib.Path:
        if self.is_zarr:
            self.temp_path = self.mkstemp_directory(
                prefix=f"{self.base_name}.",
                dir=self.dir_path,
            )
        else:
            fd, temp_filename = tempfile.mkstemp(
                prefix=f"{self.base_name}.",
                dir=self.dir_path,
            )
            os.close(fd)
            self.temp_path = pathlib.Path(temp_filename)
        return self.temp_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Clean up temporary file/directory on exception
        if exc_type is not None:
            if self.temp_path:
                if self.temp_path.is_dir():
                    shutil.rmtree(self.temp_path, ignore_errors=True)
                elif self.temp_path.exists():
                    self.temp_path.unlink(missing_ok=True)
            return False

        if self.is_zarr:
            if self.output_path.exists():
                shutil.rmtree(self.output_path)
            self.temp_path.rename(self.output_path)
        else:
            shutil.move(str(self.temp_path), str(self.output_path))

        return False

    def mkstemp_directory(
        self, prefix: str, dir: pathlib.Path
    ) -> pathlib.Path:
        """
        Retrying to avoid collision.
        """
        for _ in range(5):
            rand_suffix = uuid.uuid4().hex[:8]
            temp_name = f"{prefix}{rand_suffix}"
            temp_path = dir / temp_name
            try:
                temp_path.mkdir(parents=True, exist_ok=False)
                return temp_path
            except FileExistsError:
                continue
        else:
            raise FileExistsError(
                "Could not create a unique temporary directory after 5 attempts"
            )



import logging
import math
from datetime import datetime
from typing import Optional, Union

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarineCoordinate,
    CopernicusMarinePart,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.models import (
    ChunkType,
    CoordinateChunking,
    DatasetChunking,
    VariableChunking,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.utils import (
    datetime_to_timestamp,
    timestamp_or_datestring_to_datetime,
)
from copernicusmarine.download_functions.subset_xarray import (
    t_axis_selection,
    x_axis_selection,
)

logger = logging.getLogger("copernicusmarine")


def _get_chunks_index_arithmetic(
    requested_value: float,
    reference_chunking_step: float,
    chunk_length: Union[int, float],
    chunk_step: Union[int, float],
) -> int:
    """
    Chunk index calculation for arithmetic chunking.
    """
    return math.floor(
        (requested_value - reference_chunking_step)
        / (chunk_length * chunk_step)
    )


def _get_chunks_index_geometric(
    requested_value: float,
    reference_chunking_step: float,
    chunk_length: Union[int, float],
    factor: Union[int, float, None],
) -> int:
    """
    Chunk index calculation for geometric chunking.
    """
    absolute_coordinate = abs(requested_value - reference_chunking_step)
    if absolute_coordinate < chunk_length:
        return 0
    if factor == 1 or factor is None:
        chunk_index = math.floor(absolute_coordinate / chunk_length)
    else:
        chunk_index = math.ceil(
            math.log(absolute_coordinate / chunk_length) / math.log(factor)
        )
    return (
        -chunk_index
        if requested_value < reference_chunking_step
        else chunk_index
    )


def _get_chunk_indexes_for_coordinate(
    coordinate: CopernicusMarineCoordinate,
    requested_minimum: Optional[float],
    requested_maximum: Optional[float],
    chunking_length: Union[int, float],
) -> tuple[int, int]:
    (
        coordinate_minimum_value,
        coordinate_maximum_value,
    ) = _get_coordinate_extreme(coordinate)
    if coordinate_minimum_value is None or coordinate_maximum_value is None:
        return 0, 0
    if (
        requested_minimum is None
        or requested_minimum < coordinate_minimum_value
    ):
        requested_minimum = coordinate_minimum_value
    if (
        requested_maximum is None
        or requested_maximum > coordinate_maximum_value
    ):
        requested_maximum = coordinate_maximum_value
    index_min = 0
    index_max = 0
    if chunking_length:
        if (
            coordinate.chunk_type is None
            and coordinate.step is None
            and coordinate.values
        ):
            sorted_values = sorted(coordinate.values)
            index_max = int(
                len(
                    [
                        i
                        for i in sorted_values
                        if requested_minimum <= i <= requested_maximum  # type: ignore
                    ]
                )
                / chunking_length
            )
            # index max is actually number of chunks -1
            index_min = 0
        elif (
            coordinate.chunk_type == ChunkType.ARITHMETIC
            or coordinate.chunk_type is None
        ):
            index_min = _get_chunks_index_arithmetic(
                requested_minimum,
                coordinate.chunk_reference_coordinate
                or coordinate_minimum_value,
                chunking_length,
                coordinate.step or 1,
            )
            index_max = _get_chunks_index_arithmetic(
                requested_maximum,
                coordinate.chunk_reference_coordinate
                or coordinate_minimum_value,
                chunking_length,
                coordinate.step or 1,
            )
        elif coordinate.chunk_type == ChunkType.GEOMETRIC:
            index_min = _get_chunks_index_geometric(
                requested_minimum,
                coordinate.chunk_reference_coordinate
                or coordinate_minimum_value,
                chunking_length,
                coordinate.chunk_geometric_factor,
            )
            index_max = _get_chunks_index_geometric(
                requested_maximum,
                coordinate.chunk_reference_coordinate
                or coordinate_minimum_value,
                chunking_length,
                coordinate.chunk_geometric_factor,
            )
    return (index_min, index_max)


def get_dataset_chunking(
    dataset_subset: SubsetRequest,
    service_name: CopernicusMarineServiceNames,
    dataset_version_part: CopernicusMarinePart,
) -> DatasetChunking:
    service = dataset_version_part.get_service_by_service_name(service_name)
    axis_coordinate_mapping = service.get_axis_coordinate_id_mapping()
    variables = dataset_subset.variables or []
    number_of_chunks = 0
    variables_chunking: dict[str, VariableChunking] = {}
    coordinate_chunking: dict[str, CoordinateChunking] = {}
    variables_to_iterate = (
        [
            variable
            for variable in service.variables
            if variable.short_name in variables
            or variable.standard_name in variables
        ]
        if variables
        else service.variables
    )
    for variable in variables_to_iterate:
        number_chunks_per_variable = 1
        number_values_per_variable: Union[float, int] = 1
        for coordinate in variable.coordinates:
            if coordinate.chunking_length:
                chunking_length = coordinate.chunking_length
            else:
                continue
            (
                requested_minimum,
                requested_maximum,
            ) = _extract_requested_min_max(
                coordinate,
                dataset_subset,
                axis_coordinate_mapping,
            )
            chunk_range = _get_chunk_indexes_for_coordinate(
                coordinate=coordinate,
                requested_minimum=requested_minimum,
                requested_maximum=requested_maximum,
                chunking_length=chunking_length,
            )
            number_chunks_coordinate = chunk_range[1] - chunk_range[0] + 1
            if coordinate.coordinate_id not in coordinate_chunking:
                coordinate_chunking[
                    coordinate.coordinate_id
                ] = CoordinateChunking(
                    coordinate_id=coordinate.coordinate_id,
                    chunking_length=chunking_length,
                    number_of_chunks=number_chunks_coordinate,
                )
            number_chunks_per_variable *= number_chunks_coordinate
            number_values_per_variable *= (
                number_chunks_coordinate * chunking_length
            )
        variables_chunking[variable.short_name] = VariableChunking(
            variable_short_name=variable.short_name,
            number_values=number_values_per_variable,
            number_chunks=number_chunks_per_variable,
            # default to 2MB as it is what is intended by ARCO producer
            chunk_size=2_000_000,
        )
        number_of_chunks += number_chunks_per_variable
    return DatasetChunking(
        chunking_per_variable=variables_chunking,
        chunking_per_coordinate=coordinate_chunking,
        number_chunks=number_of_chunks,
    )


def _extract_requested_min_max(
    coordinate: CopernicusMarineCoordinate,
    subset_request: SubsetRequest,
    axis_coordinate_id_mapping: dict[str, str],
) -> tuple[Optional[float], Optional[float]]:
    if coordinate.coordinate_id in axis_coordinate_id_mapping.get("t", ""):
        temporal_selection = t_axis_selection(
            subset_request.get_temporal_parameters(
                axis_coordinate_id_mapping=axis_coordinate_id_mapping
            )
        )
        min_time = None
        max_time = None
        if isinstance(temporal_selection, slice):
            min_time_datetime = temporal_selection.start
            max_time_datetime = temporal_selection.stop
        elif isinstance(temporal_selection, datetime):
            min_time_datetime = temporal_selection
            max_time_datetime = temporal_selection
        else:
            return None, None
        if min_time_datetime:
            min_time = datetime_to_timestamp(min_time_datetime)
        if max_time_datetime:
            max_time = datetime_to_timestamp(max_time_datetime)

        return min_time, max_time
    if coordinate.coordinate_id in axis_coordinate_id_mapping.get("y", ""):
        return (
            subset_request.minimum_y,
            subset_request.maximum_y,
        )
    if coordinate.coordinate_id in axis_coordinate_id_mapping.get("x", ""):
        x_selection, _ = x_axis_selection(
            subset_request.get_geographical_parameters(
                axis_coordinate_id_mapping=axis_coordinate_id_mapping
            ).x_axis_parameters
        )
        if isinstance(x_selection, slice):
            return x_selection.start, x_selection.stop
        else:
            return (None, None)
    if coordinate.coordinate_id in axis_coordinate_id_mapping.get("z", ""):
        return subset_request.minimum_depth, subset_request.maximum_depth
    return None, None


def _get_coordinate_extreme(
    coordinate: CopernicusMarineCoordinate,
) -> tuple[Union[int, float, None], Union[int, float, None]]:
    """
    Get the extreme value of a coordinate.
    """
    coordinate_minimum_value: Union[int, float]
    coordinate_maximum_value: Union[int, float]
    if isinstance(coordinate.minimum_value, str):
        coordinate_minimum_value = float(
            timestamp_or_datestring_to_datetime(
                coordinate.minimum_value
            ).timestamp()
            * 1e3
        )
    elif coordinate.minimum_value is not None:
        coordinate_minimum_value = coordinate.minimum_value
    elif coordinate.values is not None:
        coordinate_minimum_value = min(coordinate.values)  # type: ignore
        if coordinate.coordinate_id == "time":
            coordinate_minimum_value = min(
                float(
                    timestamp_or_datestring_to_datetime(value).timestamp()
                    * 1e3
                )
                for value in coordinate.values
            )

    else:
        logger.debug(
            f"Not enough information to get chunking information"
            f"for {coordinate.coordinate_id}."
        )
        return None, None
    if isinstance(coordinate.maximum_value, str):
        coordinate_maximum_value = float(
            timestamp_or_datestring_to_datetime(
                coordinate.maximum_value
            ).timestamp()
            * 1e3
        )
    elif coordinate.maximum_value is not None:
        coordinate_maximum_value = coordinate.maximum_value
    elif coordinate.values is not None:
        coordinate_maximum_value = max(coordinate.values)  # type: ignore
        if coordinate.coordinate_id == "time":
            coordinate_maximum_value = max(
                float(
                    timestamp_or_datestring_to_datetime(value).timestamp()
                    * 1e3
                )
                for value in coordinate.values
            )

    else:
        logger.debug(
            f"Not enough information to get chunking information"
            f"for {coordinate.coordinate_id}."
        )
        return None, None
    return (
        coordinate_minimum_value,
        coordinate_maximum_value,
    )



import logging
from copy import deepcopy
from datetime import datetime
from typing import Optional, Union

import pandas as pd
from dateutil.tz import UTC
from tqdm import tqdm

from copernicusmarine.catalogue_parser.models import (
    CopernicusMarinePart,
    CopernicusMarineService,
    CopernicusMarineServiceFormat,
    CopernicusMarineServiceNames,
)
from copernicusmarine.core_functions.exceptions import (
    FormatNotSupported,
    ServiceNotSupported,
)
from copernicusmarine.core_functions.models import (
    CoordinatesSelectionMethod,
    ResponseSubset,
    SplitOnTimeOption,
)
from copernicusmarine.core_functions.request_structure import SubsetRequest
from copernicusmarine.core_functions.services_utils import RetrievalService
from copernicusmarine.core_functions.subset import (
    download_zarr_or_sparse,
    retrieve_metadata_and_check_request,
)
from copernicusmarine.core_functions.utils import (
    run_multiprocessors,
    timestamp_or_datestring_to_datetime,
)
from copernicusmarine.download_functions.utils import (
    build_filename_from_request,
)

logger = logging.getLogger("copernicusmarine")


def subset_split_on_function(
    on_variables: bool,
    on_time: Optional[SplitOnTimeOption],
    subset_request: SubsetRequest,
    concurrent_processes: Optional[int],
) -> list[ResponseSubset]:
    if not on_variables and not on_time:
        raise ValueError(
            "Split on should be requested either on variables or on time."
        )

    retrieval_service = retrieve_metadata_and_check_request(subset_request)
    if retrieval_service.service_name not in [
        CopernicusMarineServiceNames.GEOSERIES,
        CopernicusMarineServiceNames.TIMESERIES,
        CopernicusMarineServiceNames.PLATFORMSERIES,
        CopernicusMarineServiceNames.OMI_ARCO,
        CopernicusMarineServiceNames.STATIC_ARCO,
    ]:
        raise ServiceNotSupported(retrieval_service.service_name)
    if (
        retrieval_service.service_format
        == CopernicusMarineServiceFormat.SQLITE
    ):
        raise FormatNotSupported(
            CopernicusMarineServiceFormat.SQLITE.value,
            "split-on",
            "subset",
        )

    time_keys: list[tuple[datetime, datetime]] = []
    variables: list[str] = []
    if on_time:
        time_keys = get_split_time_keys_from_metadata(
            part=retrieval_service.dataset_part,
            time_frequence=on_time,
            requested_minimum_time=subset_request.start_datetime,
            requested_maximum_time=subset_request.end_datetime,
            coordinate_selection_method=subset_request.coordinates_selection_method,
        )
    if on_variables:
        variables = get_split_variable_keys_from_metadata(
            service=retrieval_service.service,
            requested_variables=set(subset_request.variables or []),
        )
    new_parameters: list[dict[str, Union[list[str], datetime]]] = []
    if time_keys and variables:
        (
            _,
            variables_with_time_dimension,
            _,
        ) = retrieval_service.dataset_part.get_coordinates().get(
            "time", (None, [], None)  # type: ignore
        )
        for variable in variables:
            if variable in variables_with_time_dimension:
                for start, end in time_keys:
                    new_parameters.append(
                        {
                            "variables": [variable],
                            "start_datetime": start,
                            "end_datetime": end,
                        }
                    )
            else:
                new_parameters.append({"variables": [variable]})
    elif time_keys:
        new_parameters = [
            {"start_datetime": start, "end_datetime": end}
            for start, end in time_keys
        ]
    elif variables:
        new_parameters = [{"variables": [var]} for var in variables]
    dataset_variables = [
        variable.short_name for variable in retrieval_service.service.variables
    ]
    download_function_parameters: list[
        tuple[SubsetRequest, RetrievalService, dict]
    ] = [
        (
            _update_output_filename(
                SubsetRequest(
                    **{
                        "disable_progress_bar": True,
                        **subset_request.model_dump(
                            exclude_unset=True,
                            exclude_defaults=True,
                            exclude_none=True,
                            exclude=set(split_on_parameter.keys()),
                        ),
                    }
                ).update(split_on_parameter),
                dataset_variables=dataset_variables,
                on_time=on_time,
                on_variables=on_variables,
                axis_coordinate_id_mapping=retrieval_service.axis_coordinate_id_mapping,
            ),
            deepcopy(retrieval_service),
            {
                "disable": subset_request.disable_progress_bar,
                "desc": name_progress_bar(split_on_parameter),
                "leave": False,
                "position": 1,
            },
        )
        for split_on_parameter in new_parameters
    ]

    responses = []
    if concurrent_processes:
        responses = run_multiprocessors(
            func=download_zarr_or_sparse,
            function_arguments=download_function_parameters,
            max_concurrent_requests=concurrent_processes,
            tdqm_bar_configuration={
                "disable": subset_request.disable_progress_bar,
                "desc": "Downloading Files",
            },
        )
    else:
        with tqdm(
            total=len(download_function_parameters),
            disable=subset_request.disable_progress_bar,
            desc="Downloading Files",
        ) as pbar:
            for download_function_parameter in download_function_parameters:
                responses.append(
                    download_zarr_or_sparse(
                        download_function_parameter[0],
                        download_function_parameter[1],
                        download_function_parameter[2],
                    )
                )
                pbar.update(1)

    return responses


def get_split_time_keys_from_metadata(
    part: CopernicusMarinePart,
    time_frequence: SplitOnTimeOption,
    requested_minimum_time: Optional[datetime],
    requested_maximum_time: Optional[datetime],
    coordinate_selection_method: CoordinatesSelectionMethod,
) -> list[tuple[datetime, datetime]]:
    time_coordinate, _, _ = part.get_coordinates().get(
        "time", (None, None, None)
    )
    if not time_coordinate:
        raise ValueError(
            "No time coordinate found in the part metadata. Cannot split on time."
        )

    values = time_coordinate.values
    if values:
        values_datetime = [
            timestamp_or_datestring_to_datetime(v) for v in values
        ]

    else:
        minimum_time = timestamp_or_datestring_to_datetime(
            time_coordinate.minimum_value  # type: ignore
        )
        maximum_time = timestamp_or_datestring_to_datetime(
            time_coordinate.maximum_value  # type: ignore
        )
        step = timestamp_or_datestring_to_datetime(
            time_coordinate.step  # type: ignore
        ) - datetime(1970, 1, 1, tzinfo=UTC)
        values_datetime = [
            minimum_time + i * step
            for i in range(int((maximum_time - minimum_time) / step) + 1)
        ]

    values_datetime.sort()

    if requested_minimum_time:
        for i, v in enumerate(values_datetime):
            if v >= requested_minimum_time:
                minimum_index = i
                break
        else:
            raise ValueError(
                f"Start datetime: {requested_minimum_time} is after "
                "the maximum time available in the dataset."
            )
        if (
            coordinate_selection_method == "outside"
            and values_datetime[minimum_index] > requested_minimum_time
        ):
            minimum_index = max(0, minimum_index - 1)
        elif coordinate_selection_method == "nearest":
            if minimum_index > 0 and abs(
                values_datetime[minimum_index - 1] - requested_minimum_time
            ) < abs(values_datetime[minimum_index] - requested_minimum_time):
                minimum_index = minimum_index - 1
    else:
        minimum_index = 0

    if requested_maximum_time:
        for i in range(len(values_datetime) - 1, -1, -1):
            if values_datetime[i] <= requested_maximum_time:
                maximum_index = i
                break
        else:
            raise ValueError(
                f"End datetime: {requested_maximum_time} is before "
                "the minimum time available in the dataset."
            )
        if (
            coordinate_selection_method == "outside"
            and values_datetime[maximum_index] < requested_maximum_time
        ):
            maximum_index = min(len(values_datetime) - 1, maximum_index + 1)
        elif coordinate_selection_method == "nearest":
            if maximum_index < len(values_datetime) - 1 and abs(
                values_datetime[maximum_index + 1] - requested_maximum_time
            ) < abs(values_datetime[maximum_index] - requested_maximum_time):
                maximum_index = maximum_index + 1
    else:
        maximum_index = len(values_datetime) - 1

    selected_values = values_datetime[minimum_index : maximum_index + 1]

    groups = group_per_frequency(
        datetimes=selected_values,
        time_frequence=time_frequence,
    )
    return [
        (group_df["time"].min(), group_df["time"].max())
        for _, group_df in groups
    ]


def group_per_frequency(
    datetimes: list[datetime],
    time_frequence: SplitOnTimeOption,
):
    df = pd.DataFrame({"time": datetimes})
    if time_frequence == "hour":
        groups = df.groupby(df["time"].dt.floor("h"))
    elif time_frequence == "day":
        groups = df.groupby(df["time"].dt.date)
    elif time_frequence == "month":
        groups = df.groupby(df["time"].apply(lambda t: (t.year, t.month)))
    elif time_frequence == "year":
        groups = df.groupby(df["time"].dt.year)
    else:
        raise ValueError(f"Invalid time option: {time_frequence}")
    return groups


def get_split_variable_keys_from_metadata(
    service: CopernicusMarineService,
    requested_variables: set[str],
) -> list[str]:
    return [
        variable.short_name
        for variable in service.variables
        if not requested_variables
        or (
            variable.short_name in requested_variables
            or variable.standard_name in requested_variables
        )
    ]


def name_progress_bar(
    split_on_parameter: dict,
) -> str:
    message = ""
    if split_on_parameter.get("start_datetime"):
        message += (
            f"{split_on_parameter['start_datetime'].strftime('%Y-%m-%d')}"
        )
    if split_on_parameter.get("end_datetime"):
        if message:
            message += " - "
        message += f"{split_on_parameter['end_datetime'].strftime('%Y-%m-%d')}"
    if split_on_parameter.get("variables"):
        if message:
            message += " - "
        message += f"{split_on_parameter['variables'][0]}"
    return message


def _update_output_filename(
    subset_request: SubsetRequest,
    on_time: Optional[SplitOnTimeOption],
    on_variables: bool,
    dataset_variables: list[str],
    axis_coordinate_id_mapping: dict[str, str],
) -> SubsetRequest:
    if subset_request.output_filename:
        parsed_filename = subset_request.output_filename.split(".")
        suffix = _get_split_on_suffix(
            subset_request,
            on_time=on_time,
            on_variables=on_variables,
        )
        if len(parsed_filename) == 1:
            return subset_request.update(
                {"output_filename": subset_request.output_filename + suffix}
            )
        else:
            return subset_request.update(
                {
                    "output_filename": (
                        ".".join(parsed_filename[:-1])
                        + suffix
                        + "."
                        + parsed_filename[-1]
                    )
                }
            )
    filename = build_filename_from_request(
        subset_request,
        subset_request.variables or dataset_variables,
        platform_ids=subset_request.platform_ids or [],
        axis_coordinate_id_mapping=axis_coordinate_id_mapping,
        time_format=_get_best_time_format(on_time) if on_time else None,
    )
    return subset_request.update({"output_filename": filename})


def _get_split_on_suffix(
    subset_request: SubsetRequest,
    on_time: Optional[SplitOnTimeOption],
    on_variables: bool,
) -> str:
    suffix = "_"
    if on_time:
        time_format = _get_best_time_format(on_time) if on_time else ""
        if subset_request.start_datetime:
            suffix += f"{subset_request.start_datetime.strftime(time_format)}"
        if subset_request.end_datetime and (
            not subset_request.start_datetime
            or subset_request.end_datetime.strftime(time_format)
            != subset_request.start_datetime.strftime(time_format)
        ):
            if suffix:
                suffix += "_"
            suffix += f"{subset_request.end_datetime.strftime(time_format)}"
    if on_variables and subset_request.variables:
        if suffix:
            suffix += "_"
        suffix += f"{subset_request.variables[0]}"
    return suffix


def _get_best_time_format(
    time_option: SplitOnTimeOption,
) -> str:
    if time_option == "hour":
        return "%Y-%m-%dT%H:%M:%S"
    elif time_option == "day":
        return "%Y-%m-%d"
    elif time_option == "month":
        return "%Y-%m"
    elif time_option == "year":
        return "%Y"



import json
import logging
from typing import Optional, Union

import click
from click import Context

from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.core_functions.click_custom_class import (
    CustomClickOptionsCommand,
)
from copernicusmarine.core_functions.documentation_utils import SUBSET_SPLIT_ON
from copernicusmarine.core_functions.fields_query_builder import (
    build_query,
    get_queryable_requested_fields,
)
from copernicusmarine.core_functions.models import (
    ResponseSubset,
    SplitOnTimeOption,
)
from copernicusmarine.core_functions.subset_split_on import (
    subset_split_on_function,
)

logger = logging.getLogger("copernicusmarine")
blank_logger = logging.getLogger("copernicusmarine_blank_logger")

DEFAULT_FIELDS_TO_INCLUDE = {
    "status",
    "message",
    "file_size",
    "data_transfer_size",
    "filename",
}


@click.command(
    cls=CustomClickOptionsCommand,
    help=SUBSET_SPLIT_ON["SPLIT_ON_DESCRIPTION_HELP"],
    short_help=SUBSET_SPLIT_ON["SPLIT_ON_SHORT_HELP"],
)
@click.option(
    "--on-variables",
    is_flag=True,
    default=False,
    help=SUBSET_SPLIT_ON["ON_VARIABLES_HELP"],
)
@click.option(
    "--on-time",
    type=click.Choice(["hour", "day", "month", "year"]),
    default=None,
    help=SUBSET_SPLIT_ON["ON_TIME_HELP"],
)
@click.option(
    "--concurrent-processes",
    type=click.IntRange(1, None),
    default=None,
    help=SUBSET_SPLIT_ON["CONCURRENT_PROCESSES_HELP"],
)
@click.pass_context
@log_exception_and_exit
def split_on(
    context: Context,
    on_variables: bool,
    on_time: Optional[SplitOnTimeOption],
    concurrent_processes: Optional[int],
):
    subset_request = context.obj.get("subset_request")
    responses = subset_split_on_function(
        on_variables=on_variables,
        on_time=on_time,
        subset_request=subset_request,
        concurrent_processes=concurrent_processes,
    )

    response_fields: Optional[str] = context.obj.get("response_fields")
    dry_run: bool = subset_request.dry_run
    if response_fields:
        fields_to_include = set(response_fields.replace(" ", "").split(","))
    elif dry_run:
        fields_to_include = {"all"}
    else:
        fields_to_include = DEFAULT_FIELDS_TO_INCLUDE

    included_fields: Optional[Union[dict, set]]
    if "all" in fields_to_include:
        included_fields = None
    elif "none" in fields_to_include:
        included_fields = set()
    else:
        queryable_fields = get_queryable_requested_fields(
            fields_to_include, ResponseSubset, "subset --response-fields"
        )
        included_fields = build_query(set(queryable_fields), ResponseSubset)

    blank_logger.info(
        json.dumps(
            [
                json.loads(
                    response.model_dump_json(
                        include=included_fields,
                        exclude_none=True,
                        exclude_unset=True,
                    )
                )
                for response in responses
            ],
            indent=2,
        )
    )
